// Code generated by 'internal/gen'. DO NOT EDIT.

package misspell

// DictBritish converts US spellings to UK spellings
var DictBritish = []string{
	"institutionalization", "institutionalisation",
	"internationalization", "internationalisation",
	"professionalization", "professionalisation",
	"compartmentalizing", "compartmentalising",
	"institutionalizing", "institutionalising",
	"internationalizing", "internationalising",
	"compartmentalized", "compartmentalised",
	"compartmentalizes", "compartmentalises",
	"decriminalization", "decriminalisation",
	"denationalization", "denationalisation",
	"fictionalizations", "fictionalisations",
	"institutionalized", "institutionalised",
	"institutionalizes", "institutionalises",
	"intellectualizing", "intellectualising",
	"internationalized", "internationalised",
	"internationalizes", "internationalises",
	"pedestrianization", "pedestrianisation",
	"professionalizing", "professionalising",
	"compartmentalize", "compartmentalise",
	"decentralization", "decentralisation",
	"demilitarization", "demilitarisation",
	"externalizations", "externalisations",
	"fictionalization", "fictionalisation",
	"institutionalize", "institutionalise",
	"intellectualized", "intellectualised",
	"intellectualizes", "intellectualises",
	"internationalize", "internationalise",
	"nationalizations", "nationalisations",
	"professionalized", "professionalised",
	"professionalizes", "professionalises",
	"rationalizations", "rationalisations",
	"sensationalizing", "sensationalising",
	"sentimentalizing", "sentimentalising",
	"acclimatization", "acclimatisation",
	"commercializing", "commercialising",
	"conceptualizing", "conceptualising",
	"contextualizing", "contextualising",
	"crystallization", "crystallisation",
	"decriminalizing", "decriminalising",
	"democratization", "democratisation",
	"denationalizing", "denationalising",
	"depersonalizing", "depersonalising",
	"desensitization", "desensitisation",
	"disorganization", "disorganisation",
	"extemporization", "extemporisation",
	"externalization", "externalisation",
	"familiarization", "familiarisation",
	"generalizations", "generalisations",
	"hospitalization", "hospitalisation",
	"individualizing", "individualising",
	"industrializing", "industrialising",
	"intellectualize", "intellectualise",
	"internalization", "internalisation",
	"maneuverability", "manoeuvrability",
	"materialization", "materialisation",
	"miniaturization", "miniaturisation",
	"nationalization", "nationalisation",
	"overemphasizing", "overemphasising",
	"paleontologists", "palaeontologists",
	"particularizing", "particularising",
	"pedestrianizing", "pedestrianising",
	"professionalize", "professionalise",
	"psychoanalyzing", "psychoanalysing",
	"rationalization", "rationalisation",
	"reorganizations", "reorganisations",
	"revolutionizing", "revolutionising",
	"sensationalized", "sensationalised",
	"sensationalizes", "sensationalises",
	"sentimentalized", "sentimentalised",
	"sentimentalizes", "sentimentalises",
	"specializations", "specialisations",
	"standardization", "standardisation",
	"synchronization", "synchronisation",
	"systematization", "systematisation",
	"aggrandizement", "aggrandisement",
	"characterizing", "characterising",
	"collectivizing", "collectivising",
	"commercialized", "commercialised",
	"commercializes", "commercialises",
	"conceptualized", "conceptualised",
	"conceptualizes", "conceptualises",
	"contextualized", "contextualised",
	"contextualizes", "contextualises",
	"decentralizing", "decentralising",
	"decriminalized", "decriminalised",
	"decriminalizes", "decriminalises",
	"dehumanization", "dehumanisation",
	"demilitarizing", "demilitarising",
	"demobilization", "demobilisation",
	"demoralization", "demoralisation",
	"denationalized", "denationalised",
	"denationalizes", "denationalises",
	"depersonalized", "depersonalised",
	"depersonalizes", "depersonalises",
	"dramatizations", "dramatisations",
	"editorializing", "editorialising",
	"fictionalizing", "fictionalising",
	"fraternization", "fraternisation",
	"generalization", "generalisation",
	"immobilization", "immobilisation",
	"individualized", "individualised",
	"individualizes", "individualises",
	"industrialized", "industrialised",
	"industrializes", "industrialises",
	"liberalization", "liberalisation",
	"monopolization", "monopolisation",
	"naturalization", "naturalisation",
	"neighborliness", "neighbourliness",
	"neutralization", "neutralisation",
	"organizational", "organisational",
	"outmaneuvering", "outmanoeuvring",
	"overemphasized", "overemphasised",
	"overemphasizes", "overemphasises",
	"paleontologist", "palaeontologist",
	"particularized", "particularised",
	"particularizes", "particularises",
	"pasteurization", "pasteurisation",
	"pedestrianized", "pedestrianised",
	"pedestrianizes", "pedestrianises",
	"philosophizing", "philosophising",
	"politicization", "politicisation",
	"popularization", "popularisation",
	"pressurization", "pressurisation",
	"prioritization", "prioritisation",
	"privatizations", "privatisations",
	"propagandizing", "propagandising",
	"psychoanalyzed", "psychoanalysed",
	"psychoanalyzes", "psychoanalyses",
	"reconnoitering", "reconnoitring",
	"regularization", "regularisation",
	"reorganization", "reorganisation",
	"revolutionized", "revolutionised",
	"revolutionizes", "revolutionises",
	"secularization", "secularisation",
	"sensationalize", "sensationalise",
	"sentimentalize", "sentimentalise",
	"serializations", "serialisations",
	"specialization", "specialisation",
	"sterilizations", "sterilisations",
	"stigmatization", "stigmatisation",
	"transistorized", "transistorised",
	"unrecognizable", "unrecognisable",
	"visualizations", "visualisations",
	"westernization", "westernisation",
	"accessorizing", "accessorising",
	"acclimatizing", "acclimatising",
	"amortizations", "amortisations",
	"amphitheaters", "amphitheatres",
	"anesthetizing", "anaesthetising",
	"archeologists", "archaeologists",
	"breathalyzers", "breathalysers",
	"breathalyzing", "breathalysing",
	"cannibalizing", "cannibalising",
	"characterized", "characterised",
	"characterizes", "characterises",
	"circularizing", "circularising",
	"collectivized", "collectivised",
	"collectivizes", "collectivises",
	"commercialize", "commercialise",
	"computerizing", "computerising",
	"conceptualize", "conceptualise",
	"contextualize", "contextualise",
	"criminalizing", "criminalising",
	"crystallizing", "crystallising",
	"decentralized", "decentralised",
	"decentralizes", "decentralises",
	"decriminalize", "decriminalise",
	"demilitarized", "demilitarised",
	"demilitarizes", "demilitarises",
	"democratizing", "democratising",
	"denationalize", "denationalise",
	"depersonalize", "depersonalise",
	"desensitizing", "desensitising",
	"destabilizing", "destabilising",
	"disemboweling", "disembowelling",
	"dramatization", "dramatisation",
	"editorialized", "editorialised",
	"editorializes", "editorialises",
	"extemporizing", "extemporising",
	"externalizing", "externalising",
	"familiarizing", "familiarising",
	"fertilization", "fertilisation",
	"fictionalized", "fictionalised",
	"fictionalizes", "fictionalises",
	"formalization", "formalisation",
	"fossilization", "fossilisation",
	"globalization", "globalisation",
	"gynecological", "gynaecological",
	"gynecologists", "gynaecologists",
	"harmonization", "harmonisation",
	"hematological", "haematological",
	"hematologists", "haematologists",
	"hospitalizing", "hospitalising",
	"hypothesizing", "hypothesising",
	"immortalizing", "immortalising",
	"individualize", "individualise",
	"industrialize", "industrialise",
	"internalizing", "internalising",
	"marginalizing", "marginalising",
	"materializing", "materialising",
	"mechanization", "mechanisation",
	"memorializing", "memorialising",
	"miniaturizing", "miniaturising",
	"nationalizing", "nationalising",
	"neighborhoods", "neighbourhoods",
	"normalization", "normalisation",
	"organizations", "organisations",
	"outmaneuvered", "outmanoeuvred",
	"overemphasize", "overemphasise",
	"particularize", "particularise",
	"passivization", "passivisation",
	"patronizingly", "patronisingly",
	"pedestrianize", "pedestrianise",
	"pediatricians", "paediatricians",
	"personalizing", "personalising",
	"philosophized", "philosophised",
	"philosophizes", "philosophises",
	"privatization", "privatisation",
	"propagandized", "propagandised",
	"propagandizes", "propagandises",
	"proselytizers", "proselytisers",
	"proselytizing", "proselytising",
	"psychoanalyze", "psychoanalyse",
	"pulverization", "pulverisation",
	"rationalizing", "rationalising",
	"reconnoitered", "reconnoitred",
	"revolutionize", "revolutionise",
	"romanticizing", "romanticising",
	"serialization", "serialisation",
	"socialization", "socialisation",
	"standardizing", "standardising",
	"sterilization", "sterilisation",
	"subsidization", "subsidisation",
	"synchronizing", "synchronising",
	"systematizing", "systematising",
	"tantalizingly", "tantalisingly",
	"underutilized", "underutilised",
	"victimization", "victimisation",
	"visualization", "visualisation",
	"vocalizations", "vocalisations",
	"vulgarization", "vulgarisation",
	"accessorized", "accessorised",
	"accessorizes", "accessorises",
	"acclimatized", "acclimatised",
	"acclimatizes", "acclimatises",
	"amortization", "amortisation",
	"amphitheater", "amphitheatre",
	"anesthetists", "anaesthetists",
	"anesthetized", "anaesthetised",
	"anesthetizes", "anaesthetises",
	"antagonizing", "antagonising",
	"appetizingly", "appetisingly",
	"archeologist", "archaeologist",
	"backpedaling", "backpedalling",
	"bastardizing", "bastardising",
	"behaviorists", "behaviourists",
	"bowdlerizing", "bowdlerising",
	"breathalyzed", "breathalysed",
	"breathalyzes", "breathalyses",
	"cannibalized", "cannibalised",
	"cannibalizes", "cannibalises",
	"capitalizing", "capitalising",
	"caramelizing", "caramelising",
	"categorizing", "categorising",
	"centerpieces", "centrepieces",
	"centralizing", "centralising",
	"characterize", "characterise",
	"circularized", "circularised",
	"circularizes", "circularises",
	"clarinetists", "clarinettists",
	"collectivize", "collectivise",
	"colonization", "colonisation",
	"computerized", "computerised",
	"computerizes", "computerises",
	"criminalized", "criminalised",
	"criminalizes", "criminalises",
	"crystallized", "crystallised",
	"crystallizes", "crystallises",
	"decentralize", "decentralise",
	"dehumanizing", "dehumanising",
	"demilitarize", "demilitarise",
	"demobilizing", "demobilising",
	"democratized", "democratised",
	"democratizes", "democratises",
	"demoralizing", "demoralising",
	"desensitized", "desensitised",
	"desensitizes", "desensitises",
	"destabilized", "destabilised",
	"destabilizes", "destabilises",
	"disemboweled", "disembowelled",
	"dishonorable", "dishonourable",
	"dishonorably", "dishonourably",
	"disorganized", "disorganised",
	"editorialize", "editorialise",
	"equalization", "equalisation",
	"evangelizing", "evangelising",
	"extemporized", "extemporised",
	"extemporizes", "extemporises",
	"externalized", "externalised",
	"externalizes", "externalises",
	"familiarized", "familiarised",
	"familiarizes", "familiarises",
	"fictionalize", "fictionalise",
	"finalization", "finalisation",
	"fraternizing", "fraternising",
	"generalizing", "generalising",
	"gynecologist", "gynaecologist",
	"hematologist", "haematologist",
	"hemophiliacs", "haemophiliacs",
	"hemorrhaging", "haemorrhaging",
	"homogenizing", "homogenising",
	"hospitalized", "hospitalised",
	"hospitalizes", "hospitalises",
	"hypothesized", "hypothesised",
	"hypothesizes", "hypothesises",
	"idealization", "idealisation",
	"immobilizers", "immobilisers",
	"immobilizing", "immobilising",
	"immortalized", "immortalised",
	"immortalizes", "immortalises",
	"immunization", "immunisation",
	"initializing", "initialising",
	"installments", "instalments",
	"internalized", "internalised",
	"internalizes", "internalises",
	"jeopardizing", "jeopardising",
	"legalization", "legalisation",
	"legitimizing", "legitimising",
	"liberalizing", "liberalising",
	"maneuverable", "manoeuvrable",
	"maneuverings", "manoeuvrings",
	"marginalized", "marginalised",
	"marginalizes", "marginalises",
	"materialized", "materialised",
	"materializes", "materialises",
	"maximization", "maximisation",
	"memorialized", "memorialised",
	"memorializes", "memorialises",
	"metabolizing", "metabolising",
	"militarizing", "militarising",
	"miniaturized", "miniaturised",
	"miniaturizes", "miniaturises",
	"miscataloged", "miscatalogued",
	"misdemeanors", "misdemeanours",
	"mobilization", "mobilisation",
	"moisturizers", "moisturisers",
	"moisturizing", "moisturising",
	"monopolizing", "monopolising",
	"multicolored", "multicoloured",
	"nationalized", "nationalised",
	"nationalizes", "nationalises",
	"naturalizing", "naturalising",
	"neighborhood", "neighbourhood",
	"neutralizing", "neutralising",
	"organization", "organisation",
	"outmaneuvers", "outmanoeuvres",
	"paleontology", "palaeontology",
	"pasteurizing", "pasteurising",
	"pediatrician", "paediatrician",
	"personalized", "personalised",
	"personalizes", "personalises",
	"philosophize", "philosophise",
	"plagiarizing", "plagiarising",
	"polarization", "polarisation",
	"politicizing", "politicising",
	"popularizing", "popularising",
	"pressurizing", "pressurising",
	"prioritizing", "prioritising",
	"propagandize", "propagandise",
	"proselytized", "proselytised",
	"proselytizer", "proselytiser",
	"proselytizes", "proselytises",
	"radicalizing", "radicalising",
	"rationalized", "rationalised",
	"rationalizes", "rationalises",
	"realizations", "realisations",
	"recognizable", "recognisable",
	"recognizably", "recognisably",
	"recognizance", "recognisance",
	"reconnoiters", "reconnoitres",
	"regularizing", "regularising",
	"reorganizing", "reorganising",
	"revitalizing", "revitalising",
	"rhapsodizing", "rhapsodising",
	"romanticized", "romanticised",
	"romanticizes", "romanticises",
	"scandalizing", "scandalising",
	"scrutinizing", "scrutinising",
	"secularizing", "secularising",
	"standardized", "standardised",
	"standardizes", "standardises",
	"stigmatizing", "stigmatising",
	"sympathizers", "sympathisers",
	"sympathizing", "sympathising",
	"synchronized", "synchronised",
	"synchronizes", "synchronises",
	"synthesizing", "synthesising",
	"systematized", "systematised",
	"systematizes", "systematises",
	"theatergoers", "theatregoers",
	"traumatizing", "traumatising",
	"trivializing", "trivialising",
	"unauthorized", "unauthorised",
	"unionization", "unionisation",
	"unrecognized", "unrecognised",
	"urbanization", "urbanisation",
	"vaporization", "vaporisation",
	"vocalization", "vocalisation",
	"westernizing", "westernising",
	"accessorize", "accessorise",
	"acclimatize", "acclimatise",
	"agonizingly", "agonisingly",
	"amortizable", "amortisable",
	"anesthetics", "anaesthetics",
	"anesthetist", "anaesthetist",
	"anesthetize", "anaesthetise",
	"anglicizing", "anglicising",
	"antagonized", "antagonised",
	"antagonizes", "antagonises",
	"apologizing", "apologising",
	"backpedaled", "backpedalled",
	"bastardized", "bastardised",
	"bastardizes", "bastardises",
	"behaviorism", "behaviourism",
	"behaviorist", "behaviourist",
	"bowdlerized", "bowdlerised",
	"bowdlerizes", "bowdlerises",
	"brutalizing", "brutalising",
	"cannibalize", "cannibalise",
	"capitalized", "capitalised",
	"capitalizes", "capitalises",
	"caramelized", "caramelised",
	"caramelizes", "caramelises",
	"carbonizing", "carbonising",
	"categorized", "categorised",
	"categorizes", "categorises",
	"cauterizing", "cauterising",
	"centerfolds", "centrefolds",
	"centerpiece", "centrepiece",
	"centiliters", "centilitres",
	"centimeters", "centimetres",
	"centralized", "centralised",
	"centralizes", "centralises",
	"circularize", "circularise",
	"clarinetist", "clarinettist",
	"computerize", "computerise",
	"criminalize", "criminalise",
	"criticizing", "criticising",
	"crystallize", "crystallise",
	"customizing", "customising",
	"defenseless", "defenceless",
	"dehumanized", "dehumanised",
	"dehumanizes", "dehumanises",
	"demobilized", "demobilised",
	"demobilizes", "demobilises",
	"democratize", "democratise",
	"demoralized", "demoralised",
	"demoralizes", "demoralises",
	"deodorizing", "deodorising",
	"desensitize", "desensitise",
	"destabilize", "destabilise",
	"discoloring", "discolouring",
	"dishonoring", "dishonouring",
	"dramatizing", "dramatising",
	"economizing", "economising",
	"empathizing", "empathising",
	"emphasizing", "emphasising",
	"endeavoring", "endeavouring",
	"epitomizing", "epitomising",
	"esophaguses", "oesophaguses",
	"evangelized", "evangelised",
	"evangelizes", "evangelises",
	"extemporize", "extemporise",
	"externalize", "externalise",
	"factorizing", "factorising",
	"familiarize", "familiarise",
	"fantasizing", "fantasising",
	"fertilizers", "fertilisers",
	"fertilizing", "fertilising",
	"formalizing", "formalising",
	"fossilizing", "fossilising",
	"fraternized", "fraternised",
	"fraternizes", "fraternises",
	"fulfillment", "fulfilment",
	"galvanizing", "galvanising",
	"generalized", "generalised",
	"generalizes", "generalises",
	"ghettoizing", "ghettoising",
	"globalizing", "globalising",
	"harmonizing", "harmonising",
	"hemophiliac", "haemophiliac",
	"hemorrhaged", "haemorrhaged",
	"hemorrhages", "haemorrhages",
	"hemorrhoids", "haemorrhoids",
	"homogenized", "homogenised",
	"homogenizes", "homogenises",
	"hospitalize", "hospitalise",
	"hybridizing", "hybridising",
	"hypnotizing", "hypnotising",
	"hypothesize", "hypothesise",
	"immobilized", "immobilised",
	"immobilizer", "immobiliser",
	"immobilizes", "immobilises",
	"immortalize", "immortalise",
	"initialized", "initialised",
	"initializes", "initialises",
	"installment", "instalment",
	"internalize", "internalise",
	"italicizing", "italicising",
	"jeopardized", "jeopardised",
	"jeopardizes", "jeopardises",
	"legitimized", "legitimised",
	"legitimizes", "legitimises",
	"liberalized", "liberalised",
	"liberalizes", "liberalises",
	"lionization", "lionisation",
	"liquidizers", "liquidisers",
	"liquidizing", "liquidising",
	"magnetizing", "magnetising",
	"maneuvering", "manoeuvring",
	"marginalize", "marginalise",
	"marvelously", "marvellously",
	"materialize", "materialise",
	"mechanizing", "mechanising",
	"memorialize", "memorialise",
	"mesmerizing", "mesmerising",
	"metabolized", "metabolised",
	"metabolizes", "metabolises",
	"militarized", "militarised",
	"militarizes", "militarises",
	"milliliters", "millilitres",
	"millimeters", "millimetres",
	"miniaturize", "miniaturise",
	"misbehavior", "misbehaviour",
	"misdemeanor", "misdemeanour",
	"modernizing", "modernising",
	"moisturized", "moisturised",
	"moisturizer", "moisturiser",
	"moisturizes", "moisturises",
	"monopolized", "monopolised",
	"monopolizes", "monopolises",
	"nationalize", "nationalise",
	"naturalized", "naturalised",
	"naturalizes", "naturalises",
	"neighboring", "neighbouring",
	"neutralized", "neutralised",
	"neutralizes", "neutralises",
	"normalizing", "normalising",
	"orthopedics", "orthopaedics",
	"ostracizing", "ostracising",
	"outmaneuver", "outmanoeuvre",
	"oxidization", "oxidisation",
	"pasteurized", "pasteurised",
	"pasteurizes", "pasteurises",
	"patronizing", "patronising",
	"personalize", "personalise",
	"plagiarized", "plagiarised",
	"plagiarizes", "plagiarises",
	"politicized", "politicised",
	"politicizes", "politicises",
	"popularized", "popularised",
	"popularizes", "popularises",
	"pressurized", "pressurised",
	"pressurizes", "pressurises",
	"prioritized", "prioritised",
	"prioritizes", "prioritises",
	"privatizing", "privatising",
	"proselytize", "proselytise",
	"publicizing", "publicising",
	"pulverizing", "pulverising",
	"radicalized", "radicalised",
	"radicalizes", "radicalises",
	"randomizing", "randomising",
	"rationalize", "rationalise",
	"realization", "realisation",
	"recognizing", "recognising",
	"reconnoiter", "reconnoitre",
	"regularized", "regularised",
	"regularizes", "regularises",
	"reorganized", "reorganised",
	"reorganizes", "reorganises",
	"revitalized", "revitalised",
	"revitalizes", "revitalises",
	"rhapsodized", "rhapsodised",
	"rhapsodizes", "rhapsodises",
	"romanticize", "romanticise",
	"scandalized", "scandalised",
	"scandalizes", "scandalises",
	"scrutinized", "scrutinised",
	"scrutinizes", "scrutinises",
	"secularized", "secularised",
	"secularizes", "secularises",
	"sensitizing", "sensitising",
	"serializing", "serialising",
	"sermonizing", "sermonising",
	"signalizing", "signalising",
	"skeptically", "sceptically",
	"socializing", "socialising",
	"solemnizing", "solemnising",
	"specialized", "specialised",
	"specializes", "specialises",
	"squirreling", "squirrelling",
	"stabilizers", "stabilisers",
	"stabilizing", "stabilising",
	"standardize", "standardise",
	"sterilizers", "sterilisers",
	"sterilizing", "sterilising",
	"stigmatized", "stigmatised",
	"stigmatizes", "stigmatises",
	"subsidizers", "subsidisers",
	"subsidizing", "subsidising",
	"summarizing", "summarising",
	"symbolizing", "symbolising",
	"sympathized", "sympathised",
	"sympathizer", "sympathiser",
	"sympathizes", "sympathises",
	"synchronize", "synchronise",
	"synthesized", "synthesised",
	"synthesizes", "synthesises",
	"systematize", "systematise",
	"tantalizing", "tantalising",
	"temporizing", "temporising",
	"tenderizing", "tenderising",
	"terrorizing", "terrorising",
	"theatergoer", "theatregoer",
	"traumatized", "traumatised",
	"traumatizes", "traumatises",
	"trivialized", "trivialised",
	"trivializes", "trivialises",
	"tyrannizing", "tyrannising",
	"uncataloged", "uncatalogued",
	"uncivilized", "uncivilised",
	"unfavorable", "unfavourable",
	"unfavorably", "unfavourably",
	"unorganized", "unorganised",
	"untrammeled", "untrammelled",
	"utilization", "utilisation",
	"vandalizing", "vandalising",
	"verbalizing", "verbalising",
	"victimizing", "victimising",
	"visualizing", "visualising",
	"vulgarizing", "vulgarising",
	"watercolors", "watercolours",
	"westernized", "westernised",
	"westernizes", "westernises",
	"amortizing", "amortising",
	"anesthesia", "anaesthesia",
	"anesthetic", "anaesthetic",
	"anglicized", "anglicised",
	"anglicizes", "anglicises",
	"annualized", "annualised",
	"antagonize", "antagonise",
	"apologized", "apologised",
	"apologizes", "apologises",
	"appetizers", "appetisers",
	"appetizing", "appetising",
	"archeology", "archaeology",
	"authorizes", "authorises",
	"bastardize", "bastardise",
	"bedeviling", "bedevilling",
	"behavioral", "behavioural",
	"belaboring", "belabouring",
	"bowdlerize", "bowdlerise",
	"brutalized", "brutalised",
	"brutalizes", "brutalises",
	"canalizing", "canalising",
	"canonizing", "canonising",
	"capitalize", "capitalise",
	"caramelize", "caramelise",
	"carbonized", "carbonised",
	"carbonizes", "carbonises",
	"cataloging", "cataloguing",
	"catalyzing", "catalysing",
	"categorize", "categorise",
	"cauterized", "cauterised",
	"cauterizes", "cauterises",
	"centerfold", "centrefold",
	"centiliter", "centilitre",
	"centimeter", "centimetre",
	"centralize", "centralise",
	"channeling", "channelling",
	"checkbooks", "chequebooks",
	"civilizing", "civilising",
	"colonizers", "colonisers",
	"colonizing", "colonising",
	"colorfully", "colourfully",
	"colorizing", "colourizing",
	"councilors", "councillors",
	"counselors", "counsellors",
	"criticized", "criticised",
	"criticizes", "criticises",
	"customized", "customised",
	"customizes", "customises",
	"dehumanize", "dehumanise",
	"demobilize", "demobilise",
	"demonizing", "demonising",
	"demoralize", "demoralise",
	"deodorized", "deodorised",
	"deodorizes", "deodorises",
	"deputizing", "deputising",
	"digitizing", "digitising",
	"discolored", "discoloured",
	"disheveled", "dishevelled",
	"dishonored", "dishonoured",
	"dramatized", "dramatised",
	"dramatizes", "dramatises",
	"economized", "economised",
	"economizes", "economises",
	"empathized", "empathised",
	"empathizes", "empathises",
	"emphasized", "emphasised",
	"emphasizes", "emphasises",
	"endeavored", "endeavoured",
	"energizing", "energising",
	"epicenters", "epicentres",
	"epitomized", "epitomised",
	"epitomizes", "epitomises",
	"equalizers", "equalisers",
	"equalizing", "equalising",
	"eulogizing", "eulogising",
	"evangelize", "evangelise",
	"factorized", "factorised",
	"factorizes", "factorises",
	"fantasized", "fantasised",
	"fantasizes", "fantasises",
	"favoritism", "favouritism",
	"feminizing", "feminising",
	"fertilized", "fertilised",
	"fertilizer", "fertiliser",
	"fertilizes", "fertilises",
	"fiberglass", "fibreglass",
	"finalizing", "finalising",
	"flavorings", "flavourings",
	"flavorless", "flavourless",
	"flavorsome", "flavoursome",
	"formalized", "formalised",
	"formalizes", "formalises",
	"fossilized", "fossilised",
	"fossilizes", "fossilises",
	"fraternize", "fraternise",
	"galvanized", "galvanised",
	"galvanizes", "galvanises",
	"generalize", "generalise",
	"ghettoized", "ghettoised",
	"ghettoizes", "ghettoises",
	"globalized", "globalised",
	"globalizes", "globalises",
	"gruelingly", "gruellingly",
	"gynecology", "gynaecology",
	"harmonized", "harmonised",
	"harmonizes", "harmonises",
	"hematology", "haematology",
	"hemoglobin", "haemoglobin",
	"hemophilia", "haemophilia",
	"hemorrhage", "haemorrhage",
	"homogenize", "homogenise",
	"humanizing", "humanising",
	"hybridized", "hybridised",
	"hybridizes", "hybridises",
	"hypnotized", "hypnotised",
	"hypnotizes", "hypnotises",
	"idealizing", "idealising",
	"immobilize", "immobilise",
	"immunizing", "immunising",
	"impaneling", "impanelling",
	"imperiling", "imperilling",
	"initialing", "initialling",
	"initialize", "initialise",
	"ionization", "ionisation",
	"italicized", "italicised",
	"italicizes", "italicises",
	"jeopardize", "jeopardise",
	"kilometers", "kilometres",
	"lackluster", "lacklustre",
	"legalizing", "legalising",
	"legitimize", "legitimise",
	"liberalize", "liberalise",
	"liquidized", "liquidised",
	"liquidizer", "liquidiser",
	"liquidizes", "liquidises",
	"localizing", "localising",
	"magnetized", "magnetised",
	"magnetizes", "magnetises",
	"maneuvered", "manoeuvred",
	"marshaling", "marshalling",
	"maximizing", "maximising",
	"mechanized", "mechanised",
	"mechanizes", "mechanises",
	"memorizing", "memorising",
	"mesmerized", "mesmerised",
	"mesmerizes", "mesmerises",
	"metabolize", "metabolise",
	"militarize", "militarise",
	"milliliter", "millilitre",
	"millimeter", "millimetre",
	"minimizing", "minimising",
	"mobilizing", "mobilising",
	"modernized", "modernised",
	"modernizes", "modernises",
	"moisturize", "moisturise",
	"monopolize", "monopolise",
	"moralizing", "moralising",
	"naturalize", "naturalise",
	"neighborly", "neighbourly",
	"neutralize", "neutralise",
	"normalized", "normalised",
	"normalizes", "normalises",
	"optimizing", "optimising",
	"organizers", "organisers",
	"organizing", "organising",
	"orthopedic", "orthopaedic",
	"ostracized", "ostracised",
	"ostracizes", "ostracises",
	"paralyzing", "paralysing",
	"pasteurize", "pasteurise",
	"patronized", "patronised",
	"patronizes", "patronises",
	"pedophiles", "paedophiles",
	"pedophilia", "paedophilia",
	"penalizing", "penalising",
	"plagiarize", "plagiarise",
	"plowshares", "ploughshares",
	"polarizing", "polarising",
	"politicize", "politicise",
	"popularize", "popularise",
	"prioritize", "prioritise",
	"privatized", "privatised",
	"privatizes", "privatises",
	"publicized", "publicised",
	"publicizes", "publicises",
	"pulverized", "pulverised",
	"pulverizes", "pulverises",
	"quarreling", "quarrelling",
	"radicalize", "radicalise",
	"randomized", "randomised",
	"randomizes", "randomises",
	"realizable", "realisable",
	"recognized", "recognised",
	"recognizes", "recognises",
	"regularize", "regularise",
	"remodeling", "remodelling",
	"reorganize", "reorganise",
	"revitalize", "revitalise",
	"rhapsodize", "rhapsodise",
	"ritualized", "ritualised",
	"sanitizing", "sanitising",
	"satirizing", "satirising",
	"scandalize", "scandalise",
	"scrutinize", "scrutinise",
	"secularize", "secularise",
	"sensitized", "sensitised",
	"sensitizes", "sensitises",
	"sepulchers", "sepulchres",
	"serialized", "serialised",
	"serializes", "serialises",
	"sermonized", "sermonised",
	"sermonizes", "sermonises",
	"shriveling", "shrivelling",
	"signalized", "signalised",
	"signalizes", "signalises",
	"skepticism", "scepticism",
	"socialized", "socialised",
	"socializes", "socialises",
	"sodomizing", "sodomising",
	"solemnized", "solemnised",
	"solemnizes", "solemnises",
	"specialize", "specialise",
	"squirreled", "squirrelled",
	"stabilized", "stabilised",
	"stabilizer", "stabiliser",
	"stabilizes", "stabilises",
	"stenciling", "stencilling",
	"sterilized", "sterilised",
	"sterilizer", "steriliser",
	"sterilizes", "sterilises",
	"stigmatize", "stigmatise",
	"subsidized", "subsidised",
	"subsidizer", "subsidiser",
	"subsidizes", "subsidises",
	"summarized", "summarised",
	"summarizes", "summarises",
	"symbolized", "symbolised",
	"symbolizes", "symbolises",
	"sympathize", "sympathise",
	"tantalized", "tantalised",
	"tantalizes", "tantalises",
	"temporized", "temporised",
	"temporizes", "temporises",
	"tenderized", "tenderised",
	"tenderizes", "tenderises",
	"terrorized", "terrorised",
	"terrorizes", "terrorises",
	"theorizing", "theorising",
	"traumatize", "traumatise",
	"trivialize", "trivialise",
	"tyrannized", "tyrannised",
	"tyrannizes", "tyrannises",
	"unionizing", "unionising",
	"unraveling", "unravelling",
	"urbanizing", "urbanising",
	"utilizable", "utilisable",
	"vandalized", "vandalised",
	"vandalizes", "vandalises",
	"vaporizing", "vaporising",
	"verbalized", "verbalised",
	"verbalizes", "verbalises",
	"victimized", "victimised",
	"victimizes", "victimises",
	"visualized", "visualised",
	"visualizes", "visualises",
	"vocalizing", "vocalising",
	"vulcanized", "vulcanised",
	"vulgarized", "vulgarised",
	"vulgarizes", "vulgarises",
	"watercolor", "watercolour",
	"westernize", "westernise",
	"womanizers", "womanisers",
	"womanizing", "womanising",
	"worshiping", "worshipping",
	"agonizing", "agonising",
	"airplanes", "aeroplanes",
	"amortized", "amortised",
	"amortizes", "amortises",
	"analyzing", "analysing",
	"apologize", "apologise",
	"appetizer", "appetiser",
	"artifacts", "artefacts",
	"baptizing", "baptising",
	"bedeviled", "bedevilled",
	"behaviors", "behaviours",
	"bejeweled", "bejewelled",
	"belabored", "belaboured",
	"brutalize", "brutalise",
	"canalized", "canalised",
	"canalizes", "canalises",
	"canonized", "canonised",
	"canonizes", "canonises",
	"carbonize", "carbonise",
	"cataloged", "catalogued",
	"catalyzed", "catalysed",
	"catalyzes", "catalyses",
	"cauterize", "cauterise",
	"channeled", "channelled",
	"checkbook", "chequebook",
	"checkered", "chequered",
	"chiseling", "chiselling",
	"civilized", "civilised",
	"civilizes", "civilises",
	"clamoring", "clamouring",
	"colonized", "colonised",
	"colonizer", "coloniser",
	"colonizes", "colonises",
	"colorants", "colourants",
	"colorized", "colourized",
	"colorizes", "colourizes",
	"colorless", "colourless",
	"councilor", "councillor",
	"counseled", "counselled",
	"counselor", "counsellor",
	"criticize", "criticise",
	"cudgeling", "cudgelling",
	"customize", "customise",
	"demonized", "demonised",
	"demonizes", "demonises",
	"deodorize", "deodorise",
	"deputized", "deputised",
	"deputizes", "deputises",
	"digitized", "digitised",
	"digitizes", "digitises",
	"discolors", "discolours",
	"dishonors", "dishonours",
	"dramatize", "dramatise",
	"driveling", "drivelling",
	"economize", "economise",
	"empathize", "empathise",
	"emphasize", "emphasise",
	"enameling", "enamelling",
	"endeavors", "endeavours",
	"energized", "energised",
	"energizes", "energises",
	"enthralls", "enthrals",
	"epicenter", "epicentre",
	"epitomize", "epitomise",
	"equalized", "equalised",
	"equalizer", "equaliser",
	"equalizes", "equalises",
	"eulogized", "eulogised",
	"eulogizes", "eulogises",
	"factorize", "factorise",
	"fantasize", "fantasise",
	"favorable", "favourable",
	"favorably", "favourably",
	"favorites", "favourites",
	"feminized", "feminised",
	"feminizes", "feminises",
	"fertilize", "fertilise",
	"finalized", "finalised",
	"finalizes", "finalises",
	"flavoring", "flavouring",
	"formalize", "formalise",
	"fossilize", "fossilise",
	"funneling", "funnelling",
	"galvanize", "galvanise",
	"gamboling", "gambolling",
	"ghettoize", "ghettoise",
	"globalize", "globalise",
	"gonorrhea", "gonorrhoea",
	"groveling", "grovelling",
	"harboring", "harbouring",
	"harmonize", "harmonise",
	"honorably", "honourably",
	"humanized", "humanised",
	"humanizes", "humanises",
	"hybridize", "hybridise",
	"hypnotize", "hypnotise",
	"idealized", "idealised",
	"idealizes", "idealises",
	"idolizing", "idolising",
	"immunized", "immunised",
	"immunizes", "immunises",
	"impaneled", "impanelled",
	"imperiled", "imperilled",
	"initialed", "initialled",
	"italicize", "italicise",
	"itemizing", "itemising",
	"kilometer", "kilometre",
	"legalized", "legalised",
	"legalizes", "legalises",
	"lionizing", "lionising",
	"liquidize", "liquidise",
	"localized", "localised",
	"localizes", "localises",
	"magnetize", "magnetise",
	"maneuvers", "manoeuvres",
	"marshaled", "marshalled",
	"marveling", "marvelling",
	"marvelous", "marvellous",
	"maximized", "maximised",
	"maximizes", "maximises",
	"mechanize", "mechanise",
	"memorized", "memorised",
	"memorizes", "memorises",
	"mesmerize", "mesmerise",
	"minimized", "minimised",
	"minimizes", "minimises",
	"mobilized", "mobilised",
	"mobilizes", "mobilises",
	"modernize", "modernise",
	"moldering", "mouldering",
	"moralized", "moralised",
	"moralizes", "moralises",
	"motorized", "motorised",
	"mustached", "moustached",
	"mustaches", "moustaches",
	"neighbors", "neighbours",
	"normalize", "normalise",
	"optimized", "optimised",
	"optimizes", "optimises",
	"organized", "organised",
	"organizer", "organiser",
	"organizes", "organises",
	"ostracize", "ostracise",
	"oxidizing", "oxidising",
	"panelists", "panellists",
	"paralyzed", "paralysed",
	"paralyzes", "paralyses",
	"parceling", "parcelling",
	"patronize", "patronise",
	"pedophile", "paedophile",
	"penalized", "penalised",
	"penalizes", "penalises",
	"penciling", "pencilling",
	"plowshare", "ploughshare",
	"polarized", "polarised",
	"polarizes", "polarises",
	"practiced", "practised",
	"pretenses", "pretences",
	"privatize", "privatise",
	"publicize", "publicise",
	"pulverize", "pulverise",
	"quarreled", "quarrelled",
	"randomize", "randomise",
	"realizing", "realising",
	"recognize", "recognise",
	"refueling", "refuelling",
	"remodeled", "remodelled",
	"remolding", "remoulding",
	"saltpeter", "saltpetre",
	"sanitized", "sanitised",
	"sanitizes", "sanitises",
	"satirized", "satirised",
	"satirizes", "satirises",
	"sensitize", "sensitise",
	"sepulcher", "sepulchre",
	"serialize", "serialise",
	"sermonize", "sermonise",
	"shoveling", "shovelling",
	"shriveled", "shrivelled",
	"signaling", "signalling",
	"signalize", "signalise",
	"skeptical", "sceptical",
	"sniveling", "snivelling",
	"snorkeled", "snorkelled",
	"socialize", "socialise",
	"sodomized", "sodomised",
	"sodomizes", "sodomises",
	"solemnize", "solemnise",
	"spiraling", "spiralling",
	"splendors", "splendours",
	"stabilize", "stabilise",
	"stenciled", "stencilled",
	"sterilize", "sterilise",
	"subsidize", "subsidise",
	"succoring", "succouring",
	"sulfurous", "sulphurous",
	"summarize", "summarise",
	"swiveling", "swivelling",
	"symbolize", "symbolise",
	"tantalize", "tantalise",
	"temporize", "temporise",
	"tenderize", "tenderise",
	"terrorize", "terrorise",
	"theorized", "theorised",
	"theorizes", "theorises",
	"travelers", "travellers",
	"traveling", "travelling",
	"tricolors", "tricolours",
	"tunneling", "tunnelling",
	"tyrannize", "tyrannise",
	"unequaled", "unequalled",
	"unionized", "unionised",
	"unionizes", "unionises",
	"unraveled", "unravelled",
	"unrivaled", "unrivalled",
	"urbanized", "urbanised",
	"urbanizes", "urbanises",
	"utilizing", "utilising",
	"vandalize", "vandalise",
	"vaporized", "vaporised",
	"vaporizes", "vaporises",
	"verbalize", "verbalise",
	"victimize", "victimise",
	"visualize", "visualise",
	"vocalized", "vocalised",
	"vocalizes", "vocalises",
	"vulgarize", "vulgarise",
	"weaseling", "weaselling",
	"womanized", "womanised",
	"womanizer", "womaniser",
	"womanizes", "womanises",
	"worshiped", "worshipped",
	"worshiper", "worshipper",
	"agonized", "agonised",
	"agonizes", "agonises",
	"airplane", "aeroplane",
	"aluminum", "aluminium",
	"amortize", "amortise",
	"analyzed", "analysed",
	"analyzes", "analyses",
	"armorers", "armourers",
	"armories", "armouries",
	"artifact", "artefact",
	"baptized", "baptised",
	"baptizes", "baptises",
	"behavior", "behaviour",
	"behooved", "behoved",
	"behooves", "behoves",
	"belabors", "belabours",
	"calibers", "calibres",
	"canalize", "canalise",
	"canonize", "canonise",
	"catalogs", "catalogues",
	"catalyze", "catalyse",
	"caviling", "cavilling",
	"centered", "centred",
	"chiseled", "chiselled",
	"civilize", "civilise",
	"clamored", "clamoured",
	"colonize", "colonise",
	"colorant", "colourant",
	"coloreds", "coloureds",
	"colorful", "colourful",
	"coloring", "colouring",
	"colorize", "colourize",
	"coziness", "cosiness",
	"cruelest", "cruellest",
	"cudgeled", "cudgelled",
	"defenses", "defences",
	"demeanor", "demeanour",
	"demonize", "demonise",
	"deputize", "deputise",
	"diarrhea", "diarrhoea",
	"digitize", "digitise",
	"disfavor", "disfavour",
	"dishonor", "dishonour",
	"distills", "distils",
	"driveled", "drivelled",
	"enameled", "enamelled",
	"enamored", "enamoured",
	"endeavor", "endeavour",
	"energize", "energise",
	"epaulets", "epaulettes",
	"equalize", "equalise",
	"estrogen", "oestrogen",
	"etiology", "aetiology",
	"eulogize", "eulogise",
	"favoring", "favouring",
	"favorite", "favourite",
	"feminize", "feminise",
	"finalize", "finalise",
	"flavored", "flavoured",
	"flutists", "flautists",
	"fulfills", "fulfils",
	"funneled", "funnelled",
	"gamboled", "gambolled",
	"graveled", "gravelled",
	"groveled", "grovelled",
	"grueling", "gruelling",
	"harbored", "harboured",
	"honoring", "honouring",
	"humanize", "humanise",
	"humoring", "humouring",
	"idealize", "idealise",
	"idolized", "idolised",
	"idolizes", "idolises",
	"immunize", "immunise",
	"ionizing", "ionising",
	"itemized", "itemised",
	"itemizes", "itemises",
	"jewelers", "jewellers",
	"labeling", "labelling",
	"laborers", "labourers",
	"laboring", "labouring",
	"legalize", "legalise",
	"leukemia", "leukaemia",
	"levelers", "levellers",
	"leveling", "levelling",
	"libeling", "libelling",
	"libelous", "libellous",
	"lionized", "lionised",
	"lionizes", "lionises",
	"localize", "localise",
	"louvered", "louvred",
	"maneuver", "manoeuvre",
	"marveled", "marvelled",
	"maximize", "maximise",
	"memorize", "memorise",
	"minimize", "minimise",
	"mobilize", "mobilise",
	"modelers", "modellers",
	"modeling", "modelling",
	"moldered", "mouldered",
	"moldiest", "mouldiest",
	"moldings", "mouldings",
	"moralize", "moralise",
	"mustache", "moustache",
	"neighbor", "neighbour",
	"odorless", "odourless",
	"offenses", "offences",
	"optimize", "optimise",
	"organize", "organise",
	"oxidized", "oxidised",
	"oxidizes", "oxidises",
	"paneling", "panelling",
	"panelist", "panellist",
	"paralyze", "paralyse",
	"parceled", "parcelled",
	"pedaling", "pedalling",
	"penalize", "penalise",
	"penciled", "pencilled",
	"polarize", "polarise",
	"pretense", "pretence",
	"pummeled", "pummelling",
	"raveling", "ravelling",
	"realized", "realised",
	"realizes", "realises",
	"refueled", "refuelled",
	"remolded", "remoulded",
	"revelers", "revellers",
	"reveling", "revelling",
	"rivaling", "rivalling",
	"sanitize", "sanitise",
	"satirize", "satirise",
	"savories", "savouries",
	"savoring", "savouring",
	"scepters", "sceptres",
	"shoveled", "shovelled",
	"signaled", "signalled",
	"skeptics", "sceptics",
	"sniveled", "snivelled",
	"sodomize", "sodomise",
	"specters", "spectres",
	"spiraled", "spiralled",
	"splendor", "splendour",
	"succored", "succoured",
	"sulfates", "sulphates",
	"sulfides", "sulphides",
	"swiveled", "swivelled",
	"tasseled", "tasselled",
	"theaters", "theatres",
	"theorize", "theorise",
	"toweling", "towelling",
	"traveler", "traveller",
	"trialing", "trialling",
	"tricolor", "tricolour",
	"tunneled", "tunnelled",
	"unionize", "unionise",
	"unsavory", "unsavoury",
	"urbanize", "urbanise",
	"utilized", "utilised",
	"utilizes", "utilises",
	"vaporize", "vaporise",
	"vocalize", "vocalise",
	"weaseled", "weaselled",
	"womanize", "womanise",
	"yodeling", "yodelling",
	"agonize", "agonise",
	"analyze", "analyse",
	"appalls", "appals",
	"armored", "armoured",
	"armorer", "armourer",
	"baptize", "baptise",
	"behoove", "behove",
	"belabor", "belabour",
	"beveled", "bevelled",
	"caliber", "calibre",
	"caroled", "carolled",
	"caviled", "cavilled",
	"centers", "centres",
	"clamors", "clamours",
	"clangor", "clangour",
	"colored", "coloured",
	"coziest", "cosiest",
	"crueler", "crueller",
	"defense", "defence",
	"dialing", "dialling",
	"dialogs", "dialogues",
	"distill", "distil",
	"dueling", "duelling",
	"enrolls", "enrols",
	"epaulet", "epaulette",
	"favored", "favoured",
	"flavors", "flavours",
	"flutist", "flautist",
	"fueling", "fuelling",
	"fulfill", "fulfil",
	"goiters", "goitres",
	"harbors", "harbours",
	"honored", "honoured",
	"humored", "humoured",
	"idolize", "idolise",
	"ionized", "ionised",
	"ionizes", "ionises",
	"itemize", "itemise",
	"jeweled", "jewelled",
	"jeweler", "jeweller",
	"jewelry", "jewellery",
	"labeled", "labelled",
	"labored", "laboured",
	"laborer", "labourer",
	"leveled", "levelled",
	"leveler", "leveller",
	"libeled", "libelled",
	"lionize", "lionise",
	"louvers", "louvres",
	"modeled", "modelled",
	"modeler", "modeller",
	"molders", "moulders",
	"moldier", "mouldier",
	"molding", "moulding",
	"molting", "moulting",
	"offense", "offence",
	"oxidize", "oxidise",
	"pajamas", "pyjamas",
	"paneled", "panelled",
	"parlors", "parlours",
	"pedaled", "pedalled",
	"plowing", "ploughing",
	"plowman", "ploughman",
	"plowmen", "ploughmen",
	"realize", "realise",
	"remolds", "remoulds",
	"reveled", "revelled",
	"reveler", "reveller",
	"rivaled", "rivalled",
	"rumored", "rumoured",
	"saviors", "saviours",
	"savored", "savoured",
	"scepter", "sceptre",
	"skeptic", "sceptic",
	"specter", "spectre",
	"succors", "succours",
	"sulfate", "sulphate",
	"sulfide", "sulphide",
	"theater", "theatre",
	"toweled", "towelled",
	"toxemia", "toxaemia",
	"trialed", "trialled",
	"utilize", "utilise",
	"yodeled", "yodelled",
	"anemia", "anaemia",
	"anemic", "anaemic",
	"appall", "appal",
	"arbors", "arbours",
	"armory", "armoury",
	"candor", "candour",
	"center", "centre",
	"clamor", "clamour",
	"colors", "colours",
	"cozier", "cosier",
	"cozies", "cosies",
	"cozily", "cosily",
	"dialed", "dialled",
	"drafty", "draughty",
	"dueled", "duelled",
	"favors", "favours",
	"fervor", "fervour",
	"fibers", "fibres",
	"flavor", "flavour",
	"fueled", "fuelled",
	"goiter", "goitre",
	"harbor", "harbour",
	"honors", "honours",
	"humors", "humours",
	"labors", "labours",
	"liters", "litres",
	"louver", "louvre",
	"luster", "lustre",
	"meager", "meagre",
	"miters", "mitres",
	"molded", "moulded",
	"molder", "moulder",
	"molted", "moulted",
	"pajama", "pyjama",
	"parlor", "parlour",
	"plowed", "ploughed",
	"rancor", "rancour",
	"remold", "remould",
	"rigors", "rigours",
	"rumors", "rumours",
	"savors", "savours",
	"savory", "savoury",
	"succor", "succour",
	"tumors", "tumours",
	"vapors", "vapours",
	"aging", "ageing",
	"arbor", "arbour",
	"ardor", "ardour",
	"armor", "armour",
	"chili", "chilli",
	"color", "colour",
	"edema", "edoema",
	"favor", "favour",
	"fecal", "faecal",
	"feces", "faeces",
	"fiber", "fibre",
	"honor", "honour",
	"humor", "humour",
	"labor", "labour",
	"liter", "litre",
	"miter", "mitre",
	"molds", "moulds",
	"moldy", "mouldy",
	"molts", "moults",
	"odors", "odours",
	"plows", "ploughs",
	"rigor", "rigour",
	"rumor", "rumour",
	"savor", "savour",
	"valor", "valour",
	"vapor", "vapour",
	"vigor", "vigour",
	"cozy", "cosy",
	"mold", "mould",
	"molt", "moult",
	"odor", "odour",
	"plow", "plough",
}
