// Code generated by 'internal/gen'. DO NOT EDIT.

package misspell

// DictAmerican converts UK spellings to US spellings
var DictAmerican = []string{
	"institutionalisation", "institutionalization",
	"internationalisation", "internationalization",
	"professionalisation", "professionalization",
	"compartmentalising", "compartmentalizing",
	"institutionalising", "institutionalizing",
	"internationalising", "internationalizing",
	"compartmentalised", "compartmentalized",
	"compartmentalises", "compartmentalizes",
	"decriminalisation", "decriminalization",
	"denationalisation", "denationalization",
	"fictionalisations", "fictionalizations",
	"institutionalised", "institutionalized",
	"institutionalises", "institutionalizes",
	"intellectualising", "intellectualizing",
	"internationalised", "internationalized",
	"internationalises", "internationalizes",
	"pedestrianisation", "pedestrianization",
	"professionalising", "professionalizing",
	"archaeologically", "archeologically",
	"compartmentalise", "compartmentalize",
	"decentralisation", "decentralization",
	"demilitarisation", "demilitarization",
	"externalisations", "externalizations",
	"fictionalisation", "fictionalization",
	"institutionalise", "institutionalize",
	"intellectualised", "intellectualized",
	"intellectualises", "intellectualizes",
	"internationalise", "internationalize",
	"nationalisations", "nationalizations",
	"palaeontologists", "paleontologists",
	"professionalised", "professionalized",
	"professionalises", "professionalizes",
	"rationalisations", "rationalizations",
	"sensationalising", "sensationalizing",
	"sentimentalising", "sentimentalizing",
	"acclimatisation", "acclimatization",
	"bougainvillaeas", "bougainvilleas",
	"commercialising", "commercializing",
	"conceptualising", "conceptualizing",
	"contextualising", "contextualizing",
	"crystallisation", "crystallization",
	"decriminalising", "decriminalizing",
	"democratisation", "democratization",
	"denationalising", "denationalizing",
	"depersonalising", "depersonalizing",
	"desensitisation", "desensitization",
	"destabilisation", "destabilization",
	"disorganisation", "disorganization",
	"extemporisation", "extemporization",
	"externalisation", "externalization",
	"familiarisation", "familiarization",
	"generalisations", "generalizations",
	"hospitalisation", "hospitalization",
	"individualising", "individualizing",
	"industrialising", "industrializing",
	"intellectualise", "intellectualize",
	"internalisation", "internalization",
	"manoeuvrability", "maneuverability",
	"marginalisation", "marginalization",
	"materialisation", "materialization",
	"miniaturisation", "miniaturization",
	"nationalisation", "nationalization",
	"neighbourliness", "neighborliness",
	"overemphasising", "overemphasizing",
	"palaeontologist", "paleontologist",
	"particularising", "particularizing",
	"pedestrianising", "pedestrianizing",
	"professionalise", "professionalize",
	"psychoanalysing", "psychoanalyzing",
	"rationalisation", "rationalization",
	"reorganisations", "reorganizations",
	"revolutionising", "revolutionizing",
	"sensationalised", "sensationalized",
	"sensationalises", "sensationalizes",
	"sentimentalised", "sentimentalized",
	"sentimentalises", "sentimentalizes",
	"specialisations", "specializations",
	"standardisation", "standardization",
	"synchronisation", "synchronization",
	"systematisation", "systematization",
	"aggrandisement", "aggrandizement",
	"anaesthetising", "anesthetizing",
	"archaeological", "archeological",
	"archaeologists", "archeologists",
	"bougainvillaea", "bougainvillea",
	"characterising", "characterizing",
	"collectivising", "collectivizing",
	"commercialised", "commercialized",
	"commercialises", "commercializes",
	"conceptualised", "conceptualized",
	"conceptualises", "conceptualizes",
	"contextualised", "contextualized",
	"contextualises", "contextualizes",
	"decentralising", "decentralizing",
	"decriminalised", "decriminalized",
	"decriminalises", "decriminalizes",
	"dehumanisation", "dehumanization",
	"demilitarising", "demilitarizing",
	"demobilisation", "demobilization",
	"demoralisation", "demoralization",
	"denationalised", "denationalized",
	"denationalises", "denationalizes",
	"depersonalised", "depersonalized",
	"depersonalises", "depersonalizes",
	"disembowelling", "disemboweling",
	"dramatisations", "dramatizations",
	"editorialising", "editorializing",
	"encyclopaedias", "encyclopedias",
	"fictionalising", "fictionalizing",
	"fraternisation", "fraternization",
	"generalisation", "generalization",
	"gynaecological", "gynecological",
	"gynaecologists", "gynecologists",
	"haematological", "hematological",
	"haematologists", "hematologists",
	"immobilisation", "immobilization",
	"individualised", "individualized",
	"individualises", "individualizes",
	"industrialised", "industrialized",
	"industrialises", "industrializes",
	"liberalisation", "liberalization",
	"monopolisation", "monopolization",
	"naturalisation", "naturalization",
	"neighbourhoods", "neighborhoods",
	"neutralisation", "neutralization",
	"organisational", "organizational",
	"outmanoeuvring", "outmaneuvering",
	"overemphasised", "overemphasized",
	"overemphasises", "overemphasizes",
	"paediatricians", "pediatricians",
	"particularised", "particularized",
	"particularises", "particularizes",
	"pasteurisation", "pasteurization",
	"pedestrianised", "pedestrianized",
	"pedestrianises", "pedestrianizes",
	"philosophising", "philosophizing",
	"politicisation", "politicization",
	"popularisation", "popularization",
	"pressurisation", "pressurization",
	"prioritisation", "prioritization",
	"privatisations", "privatizations",
	"propagandising", "propagandizing",
	"psychoanalysed", "psychoanalyzed",
	"psychoanalyses", "psychoanalyzes",
	"regularisation", "regularization",
	"reorganisation", "reorganization",
	"revolutionised", "revolutionized",
	"revolutionises", "revolutionizes",
	"secularisation", "secularization",
	"sensationalise", "sensationalize",
	"sentimentalise", "sentimentalize",
	"serialisations", "serializations",
	"specialisation", "specialization",
	"sterilisations", "sterilizations",
	"stigmatisation", "stigmatization",
	"transistorised", "transistorized",
	"unrecognisable", "unrecognizable",
	"visualisations", "visualizations",
	"westernisation", "westernization",
	"accessorising", "accessorizing",
	"acclimatising", "acclimatizing",
	"amortisations", "amortizations",
	"amphitheatres", "amphitheaters",
	"anaesthetised", "anesthetized",
	"anaesthetises", "anesthetizes",
	"anaesthetists", "anesthetists",
	"archaeologist", "archeologist",
	"backpedalling", "backpedaling",
	"behaviourists", "behaviorists",
	"breathalysers", "breathalyzers",
	"breathalysing", "breathalyzing",
	"callisthenics", "calisthenics",
	"cannibalising", "cannibalizing",
	"characterised", "characterized",
	"characterises", "characterizes",
	"circularising", "circularizing",
	"clarinettists", "clarinetists",
	"collectivised", "collectivized",
	"collectivises", "collectivizes",
	"commercialise", "commercialize",
	"computerising", "computerizing",
	"conceptualise", "conceptualize",
	"contextualise", "contextualize",
	"criminalising", "criminalizing",
	"crystallising", "crystallizing",
	"decentralised", "decentralized",
	"decentralises", "decentralizes",
	"decriminalise", "decriminalize",
	"demilitarised", "demilitarized",
	"demilitarises", "demilitarizes",
	"democratising", "democratizing",
	"denationalise", "denationalize",
	"depersonalise", "depersonalize",
	"desensitising", "desensitizing",
	"destabilising", "destabilizing",
	"disembowelled", "disemboweled",
	"dishonourable", "dishonorable",
	"dishonourably", "dishonorably",
	"dramatisation", "dramatization",
	"editorialised", "editorialized",
	"editorialises", "editorializes",
	"encyclopaedia", "encyclopedia",
	"encyclopaedic", "encyclopedic",
	"extemporising", "extemporizing",
	"externalising", "externalizing",
	"familiarising", "familiarizing",
	"fertilisation", "fertilization",
	"fictionalised", "fictionalized",
	"fictionalises", "fictionalizes",
	"formalisation", "formalization",
	"fossilisation", "fossilization",
	"globalisation", "globalization",
	"gynaecologist", "gynecologist",
	"haematologist", "hematologist",
	"haemophiliacs", "hemophiliacs",
	"haemorrhaging", "hemorrhaging",
	"harmonisation", "harmonization",
	"hospitalising", "hospitalizing",
	"hypothesising", "hypothesizing",
	"immortalising", "immortalizing",
	"individualise", "individualize",
	"industrialise", "industrialize",
	"internalising", "internalizing",
	"marginalising", "marginalizing",
	"materialising", "materializing",
	"mechanisation", "mechanization",
	"memorialising", "memorializing",
	"miniaturising", "miniaturizing",
	"miscatalogued", "miscataloged",
	"misdemeanours", "misdemeanors",
	"multicoloured", "multicolored",
	"nationalising", "nationalizing",
	"neighbourhood", "neighborhood",
	"normalisation", "normalization",
	"organisations", "organizations",
	"outmanoeuvred", "outmaneuvered",
	"outmanoeuvres", "outmaneuvers",
	"overemphasise", "overemphasize",
	"paediatrician", "pediatrician",
	"palaeontology", "paleontology",
	"particularise", "particularize",
	"passivisation", "passivization",
	"patronisingly", "patronizingly",
	"pedestrianise", "pedestrianize",
	"personalising", "personalizing",
	"philosophised", "philosophized",
	"philosophises", "philosophizes",
	"privatisation", "privatization",
	"propagandised", "propagandized",
	"propagandises", "propagandizes",
	"proselytisers", "proselytizers",
	"proselytising", "proselytizing",
	"psychoanalyse", "psychoanalyze",
	"pulverisation", "pulverization",
	"rationalising", "rationalizing",
	"reconnoitring", "reconnoitering",
	"revolutionise", "revolutionize",
	"romanticising", "romanticizing",
	"serialisation", "serialization",
	"socialisation", "socialization",
	"stabilisation", "stabilization",
	"standardising", "standardizing",
	"sterilisation", "sterilization",
	"subsidisation", "subsidization",
	"synchronising", "synchronizing",
	"systematising", "systematizing",
	"tantalisingly", "tantalizingly",
	"underutilised", "underutilized",
	"victimisation", "victimization",
	"visualisation", "visualization",
	"vocalisations", "vocalizations",
	"vulgarisation", "vulgarization",
	"accessorised", "accessorized",
	"accessorises", "accessorizes",
	"acclimatised", "acclimatized",
	"acclimatises", "acclimatizes",
	"amortisation", "amortization",
	"amphitheatre", "amphitheater",
	"anaesthetics", "anesthetics",
	"anaesthetise", "anesthetize",
	"anaesthetist", "anesthetist",
	"antagonising", "antagonizing",
	"appetisingly", "appetizingly",
	"backpedalled", "backpedaled",
	"bastardising", "bastardizing",
	"behaviourism", "behaviorism",
	"behaviourist", "behaviorist",
	"bowdlerising", "bowdlerizing",
	"breathalysed", "breathalyzed",
	"breathalyser", "breathalyzer",
	"breathalyses", "breathalyzes",
	"cannibalised", "cannibalized",
	"cannibalises", "cannibalizes",
	"capitalising", "capitalizing",
	"caramelising", "caramelizing",
	"categorising", "categorizing",
	"centigrammes", "centigrams",
	"centralising", "centralizing",
	"centrepieces", "centerpieces",
	"characterise", "characterize",
	"circularised", "circularized",
	"circularises", "circularizes",
	"clarinettist", "clarinetist",
	"collectivise", "collectivize",
	"colonisation", "colonization",
	"computerised", "computerized",
	"computerises", "computerizes",
	"criminalised", "criminalized",
	"criminalises", "criminalizes",
	"crystallised", "crystallized",
	"crystallises", "crystallizes",
	"decentralise", "decentralize",
	"dehumanising", "dehumanizing",
	"demilitarise", "demilitarize",
	"demobilising", "demobilizing",
	"democratised", "democratized",
	"democratises", "democratizes",
	"demoralising", "demoralizing",
	"desensitised", "desensitized",
	"desensitises", "desensitizes",
	"destabilised", "destabilized",
	"destabilises", "destabilizes",
	"discolouring", "discoloring",
	"dishonouring", "dishonoring",
	"disorganised", "disorganized",
	"editorialise", "editorialize",
	"endeavouring", "endeavoring",
	"equalisation", "equalization",
	"evangelising", "evangelizing",
	"extemporised", "extemporized",
	"extemporises", "extemporizes",
	"externalised", "externalized",
	"externalises", "externalizes",
	"familiarised", "familiarized",
	"familiarises", "familiarizes",
	"fictionalise", "fictionalize",
	"finalisation", "finalization",
	"fraternising", "fraternizing",
	"generalising", "generalizing",
	"haemophiliac", "hemophiliac",
	"haemorrhaged", "hemorrhaged",
	"haemorrhages", "hemorrhages",
	"haemorrhoids", "hemorrhoids",
	"homoeopathic", "homeopathic",
	"homogenising", "homogenizing",
	"hospitalised", "hospitalized",
	"hospitalises", "hospitalizes",
	"hypothesised", "hypothesized",
	"hypothesises", "hypothesizes",
	"idealisation", "idealization",
	"immobilisers", "immobilizers",
	"immobilising", "immobilizing",
	"immortalised", "immortalized",
	"immortalises", "immortalizes",
	"immunisation", "immunization",
	"initialising", "initializing",
	"internalised", "internalized",
	"internalises", "internalizes",
	"jeopardising", "jeopardizing",
	"legalisation", "legalization",
	"legitimising", "legitimizing",
	"liberalising", "liberalizing",
	"manoeuvrable", "maneuverable",
	"manoeuvrings", "maneuverings",
	"marginalised", "marginalized",
	"marginalises", "marginalizes",
	"marvellously", "marvelously",
	"materialised", "materialized",
	"materialises", "materializes",
	"maximisation", "maximization",
	"memorialised", "memorialized",
	"memorialises", "memorializes",
	"metabolising", "metabolizing",
	"militarising", "militarizing",
	"milligrammes", "milligrams",
	"miniaturised", "miniaturized",
	"miniaturises", "miniaturizes",
	"misbehaviour", "misbehavior",
	"misdemeanour", "misdemeanor",
	"mobilisation", "mobilization",
	"moisturisers", "moisturizers",
	"moisturising", "moisturizing",
	"monopolising", "monopolizing",
	"moustachioed", "mustachioed",
	"nationalised", "nationalized",
	"nationalises", "nationalizes",
	"naturalising", "naturalizing",
	"neighbouring", "neighboring",
	"neutralising", "neutralizing",
	"oesophaguses", "esophaguses",
	"organisation", "organization",
	"orthopaedics", "orthopedics",
	"outmanoeuvre", "outmaneuver",
	"palaeolithic", "paleolithic",
	"pasteurising", "pasteurizing",
	"personalised", "personalized",
	"personalises", "personalizes",
	"philosophise", "philosophize",
	"plagiarising", "plagiarizing",
	"ploughshares", "plowshares",
	"polarisation", "polarization",
	"politicising", "politicizing",
	"popularising", "popularizing",
	"pressurising", "pressurizing",
	"prioritising", "prioritizing",
	"propagandise", "propagandize",
	"proselytised", "proselytized",
	"proselytiser", "proselytizer",
	"proselytises", "proselytizes",
	"radicalising", "radicalizing",
	"rationalised", "rationalized",
	"rationalises", "rationalizes",
	"realisations", "realizations",
	"recognisable", "recognizable",
	"recognisably", "recognizably",
	"recognisance", "recognizance",
	"reconnoitred", "reconnoitered",
	"reconnoitres", "reconnoiters",
	"regularising", "regularizing",
	"reorganising", "reorganizing",
	"revitalising", "revitalizing",
	"rhapsodising", "rhapsodizing",
	"romanticised", "romanticized",
	"romanticises", "romanticizes",
	"scandalising", "scandalizing",
	"scrutinising", "scrutinizing",
	"secularising", "secularizing",
	"specialising", "specializing",
	"squirrelling", "squirreling",
	"standardised", "standardized",
	"standardises", "standardizes",
	"stigmatising", "stigmatizing",
	"sympathisers", "sympathizers",
	"sympathising", "sympathizing",
	"synchronised", "synchronized",
	"synchronises", "synchronizes",
	"synthesisers", "synthesizers",
	"synthesising", "synthesizing",
	"systematised", "systematized",
	"systematises", "systematizes",
	"technicolour", "technicolor",
	"theatregoers", "theatergoers",
	"traumatising", "traumatizing",
	"trivialising", "trivializing",
	"unauthorised", "unauthorized",
	"uncatalogued", "uncataloged",
	"unfavourable", "unfavorable",
	"unfavourably", "unfavorably",
	"unionisation", "unionization",
	"unrecognised", "unrecognized",
	"untrammelled", "untrammeled",
	"urbanisation", "urbanization",
	"vaporisation", "vaporization",
	"vocalisation", "vocalization",
	"watercolours", "watercolors",
	"westernising", "westernizing",
	"accessorise", "accessorize",
	"acclimatise", "acclimatize",
	"agonisingly", "agonizingly",
	"amortisable", "amortizable",
	"anaesthesia", "anesthesia",
	"anaesthetic", "anesthetic",
	"anglicising", "anglicizing",
	"antagonised", "antagonized",
	"antagonises", "antagonizes",
	"apologising", "apologizing",
	"archaeology", "archeology",
	"authorising", "authorizing",
	"bastardised", "bastardized",
	"bastardises", "bastardizes",
	"bedevilling", "bedeviling",
	"behavioural", "behavioral",
	"belabouring", "belaboring",
	"bowdlerised", "bowdlerized",
	"bowdlerises", "bowdlerizes",
	"breathalyse", "breathalyze",
	"brutalising", "brutalizing",
	"cannibalise", "cannibalize",
	"capitalised", "capitalized",
	"capitalises", "capitalizes",
	"caramelised", "caramelized",
	"caramelises", "caramelizes",
	"carbonising", "carbonizing",
	"cataloguing", "cataloging",
	"categorised", "categorized",
	"categorises", "categorizes",
	"cauterising", "cauterizing",
	"centigramme", "centigram",
	"centilitres", "centiliters",
	"centimetres", "centimeters",
	"centralised", "centralized",
	"centralises", "centralizes",
	"centrefolds", "centerfolds",
	"centrepiece", "centerpiece",
	"channelling", "channeling",
	"chequebooks", "checkbooks",
	"circularise", "circularize",
	"colourfully", "colorfully",
	"colourizing", "colorizing",
	"computerise", "computerize",
	"councillors", "councilors",
	"counselling", "counseling",
	"counsellors", "counselors",
	"criminalise", "criminalize",
	"criticising", "criticizing",
	"crystallise", "crystallize",
	"customising", "customizing",
	"defenceless", "defenseless",
	"dehumanised", "dehumanized",
	"dehumanises", "dehumanizes",
	"demobilised", "demobilized",
	"demobilises", "demobilizes",
	"democratise", "democratize",
	"demoralised", "demoralized",
	"demoralises", "demoralizes",
	"deodorising", "deodorizing",
	"desensitise", "desensitize",
	"destabilise", "destabilize",
	"discoloured", "discolored",
	"dishevelled", "disheveled",
	"dishonoured", "dishonored",
	"dramatising", "dramatizing",
	"economising", "economizing",
	"empathising", "empathizing",
	"emphasising", "emphasizing",
	"endeavoured", "endeavored",
	"epitomising", "epitomizing",
	"evangelised", "evangelized",
	"evangelises", "evangelizes",
	"extemporise", "extemporize",
	"externalise", "externalize",
	"factorising", "factorizing",
	"familiarise", "familiarize",
	"fantasising", "fantasizing",
	"favouritism", "favoritism",
	"fertilisers", "fertilizers",
	"fertilising", "fertilizing",
	"flavourings", "flavorings",
	"flavourless", "flavorless",
	"flavoursome", "flavorsome",
	"formalising", "formalizing",
	"fossilising", "fossilizing",
	"fraternised", "fraternized",
	"fraternises", "fraternizes",
	"galvanising", "galvanizing",
	"generalised", "generalized",
	"generalises", "generalizes",
	"ghettoising", "ghettoizing",
	"globalising", "globalizing",
	"gruellingly", "gruelingly",
	"gynaecology", "gynecology",
	"haematology", "hematology",
	"haemoglobin", "hemoglobin",
	"haemophilia", "hemophilia",
	"haemorrhage", "hemorrhage",
	"harmonising", "harmonizing",
	"homoeopaths", "homeopaths",
	"homoeopathy", "homeopathy",
	"homogenised", "homogenized",
	"homogenises", "homogenizes",
	"hospitalise", "hospitalize",
	"hybridising", "hybridizing",
	"hypnotising", "hypnotizing",
	"hypothesise", "hypothesize",
	"immobilised", "immobilized",
	"immobiliser", "immobilizer",
	"immobilises", "immobilizes",
	"immortalise", "immortalize",
	"impanelling", "impaneling",
	"imperilling", "imperiling",
	"initialised", "initialized",
	"initialises", "initializes",
	"initialling", "initialing",
	"instalments", "installments",
	"internalise", "internalize",
	"italicising", "italicizing",
	"jeopardised", "jeopardized",
	"jeopardises", "jeopardizes",
	"kilogrammes", "kilograms",
	"legitimised", "legitimized",
	"legitimises", "legitimizes",
	"liberalised", "liberalized",
	"liberalises", "liberalizes",
	"lionisation", "lionization",
	"liquidisers", "liquidizers",
	"liquidising", "liquidizing",
	"magnetising", "magnetizing",
	"manoeuvring", "maneuvering",
	"marginalise", "marginalize",
	"marshalling", "marshaling",
	"materialise", "materialize",
	"mechanising", "mechanizing",
	"memorialise", "memorialize",
	"mesmerising", "mesmerizing",
	"metabolised", "metabolized",
	"metabolises", "metabolizes",
	"micrometres", "micrometers",
	"militarised", "militarized",
	"militarises", "militarizes",
	"milligramme", "milligram",
	"millilitres", "milliliters",
	"millimetres", "millimeters",
	"miniaturise", "miniaturize",
	"modernising", "modernizing",
	"moisturised", "moisturized",
	"moisturiser", "moisturizer",
	"moisturises", "moisturizes",
	"monopolised", "monopolized",
	"monopolises", "monopolizes",
	"nationalise", "nationalize",
	"naturalised", "naturalized",
	"naturalises", "naturalizes",
	"neighbourly", "neighborly",
	"neutralised", "neutralized",
	"neutralises", "neutralizes",
	"normalising", "normalizing",
	"orthopaedic", "orthopedic",
	"ostracising", "ostracizing",
	"oxidisation", "oxidization",
	"paediatrics", "pediatrics",
	"paedophiles", "pedophiles",
	"paedophilia", "pedophilia",
	"passivising", "passivizing",
	"pasteurised", "pasteurized",
	"pasteurises", "pasteurizes",
	"patronising", "patronizing",
	"personalise", "personalize",
	"plagiarised", "plagiarized",
	"plagiarises", "plagiarizes",
	"ploughshare", "plowshare",
	"politicised", "politicized",
	"politicises", "politicizes",
	"popularised", "popularized",
	"popularises", "popularizes",
	"praesidiums", "presidiums",
	"pressurised", "pressurized",
	"pressurises", "pressurizes",
	"prioritised", "prioritized",
	"prioritises", "prioritizes",
	"privatising", "privatizing",
	"proselytise", "proselytize",
	"publicising", "publicizing",
	"pulverising", "pulverizing",
	"quarrelling", "quarreling",
	"radicalised", "radicalized",
	"radicalises", "radicalizes",
	"randomising", "randomizing",
	"rationalise", "rationalize",
	"realisation", "realization",
	"recognising", "recognizing",
	"reconnoitre", "reconnoiter",
	"regularised", "regularized",
	"regularises", "regularizes",
	"remodelling", "remodeling",
	"reorganised", "reorganized",
	"reorganises", "reorganizes",
	"revitalised", "revitalized",
	"revitalises", "revitalizes",
	"rhapsodised", "rhapsodized",
	"rhapsodises", "rhapsodizes",
	"romanticise", "romanticize",
	"scandalised", "scandalized",
	"scandalises", "scandalizes",
	"sceptically", "skeptically",
	"scrutinised", "scrutinized",
	"scrutinises", "scrutinizes",
	"secularised", "secularized",
	"secularises", "secularizes",
	"sensitising", "sensitizing",
	"serialising", "serializing",
	"sermonising", "sermonizing",
	"shrivelling", "shriveling",
	"signalising", "signalizing",
	"snorkelling", "snorkeling",
	"snowploughs", "snowplow",
	"socialising", "socializing",
	"solemnising", "solemnizing",
	"specialised", "specialized",
	"specialises", "specializes",
	"squirrelled", "squirreled",
	"stabilisers", "stabilizers",
	"stabilising", "stabilizing",
	"standardise", "standardize",
	"stencilling", "stenciling",
	"sterilisers", "sterilizers",
	"sterilising", "sterilizing",
	"stigmatised", "stigmatized",
	"stigmatises", "stigmatizes",
	"subsidisers", "subsidizers",
	"subsidising", "subsidizing",
	"summarising", "summarizing",
	"symbolising", "symbolizing",
	"sympathised", "sympathized",
	"sympathiser", "sympathizer",
	"sympathises", "sympathizes",
	"synchronise", "synchronize",
	"synthesised", "synthesized",
	"synthesiser", "synthesizer",
	"synthesises", "synthesizes",
	"systematise", "systematize",
	"tantalising", "tantalizing",
	"temporising", "temporizing",
	"tenderising", "tenderizing",
	"terrorising", "terrorizing",
	"theatregoer", "theatergoer",
	"traumatised", "traumatized",
	"traumatises", "traumatizes",
	"trivialised", "trivialized",
	"trivialises", "trivializes",
	"tyrannising", "tyrannizing",
	"uncivilised", "uncivilized",
	"unorganised", "unorganized",
	"unravelling", "unraveling",
	"utilisation", "utilization",
	"vandalising", "vandalizing",
	"verbalising", "verbalizing",
	"victimising", "victimizing",
	"visualising", "visualizing",
	"vulgarising", "vulgarizing",
	"watercolour", "watercolor",
	"westernised", "westernized",
	"westernises", "westernizes",
	"worshipping", "worshiping",
	"aeroplanes", "airplanes",
	"amortising", "amortizing",
	"anglicised", "anglicized",
	"anglicises", "anglicizes",
	"annualised", "annualized",
	"antagonise", "antagonize",
	"apologised", "apologized",
	"apologises", "apologizes",
	"appetisers", "appetizers",
	"appetising", "appetizing",
	"authorised", "authorized",
	"authorises", "authorizes",
	"bannisters", "banisters",
	"bastardise", "bastardize",
	"bedevilled", "bedeviled",
	"behaviours", "behaviors",
	"bejewelled", "bejeweled",
	"belaboured", "belabored",
	"bowdlerise", "bowdlerize",
	"brutalised", "brutalized",
	"brutalises", "brutalizes",
	"canalising", "canalizing",
	"cancelling", "canceling",
	"canonising", "canonizing",
	"capitalise", "capitalize",
	"caramelise", "caramelize",
	"carbonised", "carbonized",
	"carbonises", "carbonizes",
	"catalogued", "cataloged",
	"catalogues", "catalogs",
	"catalysing", "catalyzing",
	"categorise", "categorize",
	"cauterised", "cauterized",
	"cauterises", "cauterizes",
	"centilitre", "centiliter",
	"centimetre", "centimeter",
	"centralise", "centralize",
	"centrefold", "centerfold",
	"channelled", "channeled",
	"chequebook", "checkbook",
	"chiselling", "chiseling",
	"civilising", "civilizing",
	"clamouring", "clamoring",
	"colonisers", "colonizers",
	"colonising", "colonizing",
	"colourants", "colorants",
	"colourized", "colorized",
	"colourizes", "colorizes",
	"colourless", "colorless",
	"connexions", "connections",
	"councillor", "councilor",
	"counselled", "counseled",
	"counsellor", "counselor",
	"criticised", "criticized",
	"criticises", "criticizes",
	"cudgelling", "cudgeling",
	"customised", "customized",
	"customises", "customizes",
	"dehumanise", "dehumanize",
	"demobilise", "demobilize",
	"demonising", "demonizing",
	"demoralise", "demoralize",
	"deodorised", "deodorized",
	"deodorises", "deodorizes",
	"deputising", "deputizing",
	"digitising", "digitizing",
	"discolours", "discolors",
	"dishonours", "dishonors",
	"dramatised", "dramatized",
	"dramatises", "dramatizes",
	"drivelling", "driveling",
	"economised", "economized",
	"economises", "economizes",
	"empathised", "empathized",
	"empathises", "empathizes",
	"emphasised", "emphasized",
	"emphasises", "emphasizes",
	"enamelling", "enameling",
	"endeavours", "endeavors",
	"energising", "energizing",
	"epaulettes", "epaulets",
	"epicentres", "epicenters",
	"epitomised", "epitomized",
	"epitomises", "epitomizes",
	"equalisers", "equalizers",
	"equalising", "equalizing",
	"eulogising", "eulogizing",
	"evangelise", "evangelize",
	"factorised", "factorized",
	"factorises", "factorizes",
	"fantasised", "fantasized",
	"fantasises", "fantasizes",
	"favourable", "favorable",
	"favourably", "favorably",
	"favourites", "favorites",
	"feminising", "feminizing",
	"fertilised", "fertilized",
	"fertiliser", "fertilizer",
	"fertilises", "fertilizes",
	"fibreglass", "fiberglass",
	"finalising", "finalizing",
	"flavouring", "flavoring",
	"formalised", "formalized",
	"formalises", "formalizes",
	"fossilised", "fossilized",
	"fossilises", "fossilizes",
	"fraternise", "fraternize",
	"fulfilment", "fulfillment",
	"funnelling", "funneling",
	"galvanised", "galvanized",
	"galvanises", "galvanizes",
	"gambolling", "gamboling",
	"gaolbreaks", "jailbreaks",
	"generalise", "generalize",
	"ghettoised", "ghettoized",
	"ghettoises", "ghettoizes",
	"globalised", "globalized",
	"globalises", "globalizes",
	"gonorrhoea", "gonorrhea",
	"grovelling", "groveling",
	"harbouring", "harboring",
	"harmonised", "harmonized",
	"harmonises", "harmonizes",
	"homoeopath", "homeopath",
	"homogenise", "homogenize",
	"honourable", "honorable",
	"honourably", "honorably",
	"humanising", "humanizing",
	"humourless", "humorless",
	"hybridised", "hybridized",
	"hybridises", "hybridizes",
	"hypnotised", "hypnotized",
	"hypnotises", "hypnotizes",
	"idealising", "idealizing",
	"immobilise", "immobilize",
	"immunising", "immunizing",
	"impanelled", "impaneled",
	"imperilled", "imperiled",
	"inflexions", "inflections",
	"initialise", "initialize",
	"initialled", "initialed",
	"instalment", "installment",
	"ionisation", "ionization",
	"italicised", "italicized",
	"italicises", "italicizes",
	"jeopardise", "jeopardize",
	"kilogramme", "kilogram",
	"kilometres", "kilometers",
	"lacklustre", "lackluster",
	"legalising", "legalizing",
	"legitimise", "legitimize",
	"liberalise", "liberalize",
	"liquidised", "liquidized",
	"liquidiser", "liquidizer",
	"liquidises", "liquidizes",
	"localising", "localizing",
	"magnetised", "magnetized",
	"magnetises", "magnetizes",
	"manoeuvred", "maneuvered",
	"manoeuvres", "maneuvers",
	"marshalled", "marshaled",
	"marvelling", "marveling",
	"marvellous", "marvelous",
	"maximising", "maximizing",
	"mechanised", "mechanized",
	"mechanises", "mechanizes",
	"memorising", "memorizing",
	"mesmerised", "mesmerized",
	"mesmerises", "mesmerizes",
	"metabolise", "metabolize",
	"micrometre", "micrometer",
	"militarise", "militarize",
	"millilitre", "milliliter",
	"millimetre", "millimeter",
	"minimising", "minimizing",
	"mobilising", "mobilizing",
	"modernised", "modernized",
	"modernises", "modernizes",
	"moisturise", "moisturize",
	"monopolise", "monopolize",
	"moralising", "moralizing",
	"mouldering", "moldering",
	"moustached", "mustached",
	"moustaches", "mustaches",
	"naturalise", "naturalize",
	"neighbours", "neighbors",
	"neutralise", "neutralize",
	"normalised", "normalized",
	"normalises", "normalizes",
	"oesophagus", "esophagus",
	"optimising", "optimizing",
	"organisers", "organizers",
	"organising", "organizing",
	"ostracised", "ostracized",
	"ostracises", "ostracizes",
	"paederasts", "pederasts",
	"paediatric", "pediatric",
	"paedophile", "pedophile",
	"panellists", "panelists",
	"paralysing", "paralyzing",
	"parcelling", "parceling",
	"passivised", "passivized",
	"passivises", "passivizes",
	"pasteurise", "pasteurize",
	"patronised", "patronized",
	"patronises", "patronizes",
	"penalising", "penalizing",
	"pencilling", "penciling",
	"plagiarise", "plagiarize",
	"polarising", "polarizing",
	"politicise", "politicize",
	"popularise", "popularize",
	"practising", "practicing",
	"praesidium", "presidium",
	"pressurise", "pressurize",
	"prioritise", "prioritize",
	"privatised", "privatized",
	"privatises", "privatizes",
	"programmes", "programs",
	"publicised", "publicized",
	"publicises", "publicizes",
	"pulverised", "pulverized",
	"pulverises", "pulverizes",
	"pummelling", "pummeled",
	"quarrelled", "quarreled",
	"radicalise", "radicalize",
	"randomised", "randomized",
	"randomises", "randomizes",
	"realisable", "realizable",
	"recognised", "recognized",
	"recognises", "recognizes",
	"refuelling", "refueling",
	"regularise", "regularize",
	"remodelled", "remodeled",
	"remoulding", "remolding",
	"reorganise", "reorganize",
	"revitalise", "revitalize",
	"rhapsodise", "rhapsodize",
	"ritualised", "ritualized",
	"sanitising", "sanitizing",
	"satirising", "satirizing",
	"scandalise", "scandalize",
	"scepticism", "skepticism",
	"scrutinise", "scrutinize",
	"secularise", "secularize",
	"sensitised", "sensitized",
	"sensitises", "sensitizes",
	"sepulchres", "sepulchers",
	"serialised", "serialized",
	"serialises", "serializes",
	"sermonised", "sermonized",
	"sermonises", "sermonizes",
	"shovelling", "shoveling",
	"shrivelled", "shriveled",
	"signalised", "signalized",
	"signalises", "signalizes",
	"signalling", "signaling",
	"snivelling", "sniveling",
	"snorkelled", "snorkeled",
	"snowplough", "snowplow",
	"socialised", "socialized",
	"socialises", "socializes",
	"sodomising", "sodomizing",
	"solemnised", "solemnized",
	"solemnises", "solemnizes",
	"specialise", "specialize",
	"spiralling", "spiraling",
	"splendours", "splendors",
	"stabilised", "stabilized",
	"stabiliser", "stabilizer",
	"stabilises", "stabilizes",
	"stencilled", "stenciled",
	"sterilised", "sterilized",
	"steriliser", "sterilizer",
	"sterilises", "sterilizes",
	"stigmatise", "stigmatize",
	"subsidised", "subsidized",
	"subsidiser", "subsidizer",
	"subsidises", "subsidizes",
	"succouring", "succoring",
	"sulphurous", "sulfurous",
	"summarised", "summarized",
	"summarises", "summarizes",
	"swivelling", "swiveling",
	"symbolised", "symbolized",
	"symbolises", "symbolizes",
	"sympathise", "sympathize",
	"synthesise", "synthesize",
	"tantalised", "tantalized",
	"tantalises", "tantalizes",
	"temporised", "temporized",
	"temporises", "temporizes",
	"tenderised", "tenderized",
	"tenderises", "tenderizes",
	"terrorised", "terrorized",
	"terrorises", "terrorizes",
	"theorising", "theorizing",
	"traumatise", "traumatize",
	"travellers", "travelers",
	"travelling", "traveling",
	"tricolours", "tricolors",
	"trivialise", "trivialize",
	"tunnelling", "tunneling",
	"tyrannised", "tyrannized",
	"tyrannises", "tyrannizes",
	"unequalled", "unequaled",
	"unionising", "unionizing",
	"unravelled", "unraveled",
	"unrivalled", "unrivaled",
	"urbanising", "urbanizing",
	"utilisable", "utilizable",
	"vandalised", "vandalized",
	"vandalises", "vandalizes",
	"vaporising", "vaporizing",
	"verbalised", "verbalized",
	"verbalises", "verbalizes",
	"victimised", "victimized",
	"victimises", "victimizes",
	"visualised", "visualized",
	"visualises", "visualizes",
	"vocalising", "vocalizing",
	"vulcanised", "vulcanized",
	"vulgarised", "vulgarized",
	"vulgarises", "vulgarizes",
	"weaselling", "weaseling",
	"westernise", "westernize",
	"womanisers", "womanizers",
	"womanising", "womanizing",
	"worshipped", "worshiped",
	"worshipper", "worshiper",
	"aeroplane", "airplane",
	"aetiology", "etiology",
	"agonising", "agonizing",
	"almanacks", "almanacs",
	"aluminium", "aluminum",
	"amortised", "amortized",
	"amortises", "amortizes",
	"analogues", "analogs",
	"analysing", "analyzing",
	"anglicise", "anglicize",
	"apologise", "apologize",
	"appetiser", "appetizer",
	"armourers", "armorers",
	"armouries", "armories",
	"artefacts", "artifacts",
	"authorise", "authorize",
	"baptising", "baptizing",
	"behaviour", "behavior",
	"belabours", "belabors",
	"brutalise", "brutalize",
	"callipers", "calipers",
	"canalised", "canalized",
	"canalises", "canalizes",
	"cancelled", "canceled",
	"canonised", "canonized",
	"canonises", "canonizes",
	"carbonise", "carbonize",
	"carolling", "caroling",
	"catalogue", "catalog",
	"catalysed", "catalyzed",
	"catalyses", "catalyzes",
	"cauterise", "cauterize",
	"cavilling", "caviling",
	"chequered", "checkered",
	"chiselled", "chiseled",
	"civilised", "civilized",
	"civilises", "civilizes",
	"clamoured", "clamored",
	"colonised", "colonized",
	"coloniser", "colonizer",
	"colonises", "colonizes",
	"colourant", "colorant",
	"coloureds", "coloreds",
	"colourful", "colorful",
	"colouring", "coloring",
	"colourize", "colorize",
	"connexion", "connection",
	"criticise", "criticize",
	"cruellest", "cruelest",
	"cudgelled", "cudgeled",
	"customise", "customize",
	"demeanour", "demeanor",
	"demonised", "demonized",
	"demonises", "demonizes",
	"deodorise", "deodorize",
	"deputised", "deputized",
	"deputises", "deputizes",
	"dialogues", "dialogs",
	"diarrhoea", "diarrhea",
	"digitised", "digitized",
	"digitises", "digitizes",
	"discolour", "discolor",
	"disfavour", "disfavor",
	"dishonour", "dishonor",
	"dramatise", "dramatize",
	"drivelled", "driveled",
	"economise", "economize",
	"empathise", "empathize",
	"emphasise", "emphasize",
	"enamelled", "enameled",
	"enamoured", "enamored",
	"endeavour", "endeavor",
	"energised", "energized",
	"energises", "energizes",
	"epaulette", "epaulet",
	"epicentre", "epicenter",
	"epitomise", "epitomize",
	"equalised", "equalized",
	"equaliser", "equalizer",
	"equalises", "equalizes",
	"eulogised", "eulogized",
	"eulogises", "eulogizes",
	"factorise", "factorize",
	"fantasise", "fantasize",
	"favouring", "favoring",
	"favourite", "favorite",
	"feminised", "feminized",
	"feminises", "feminizes",
	"fertilise", "fertilize",
	"finalised", "finalized",
	"finalises", "finalizes",
	"flautists", "flutists",
	"flavoured", "flavored",
	"formalise", "formalize",
	"fossilise", "fossilize",
	"funnelled", "funneled",
	"galvanise", "galvanize",
	"gambolled", "gamboled",
	"gaolbirds", "jailbirds",
	"gaolbreak", "jailbreak",
	"ghettoise", "ghettoize",
	"globalise", "globalize",
	"gravelled", "graveled",
	"grovelled", "groveled",
	"gruelling", "grueling",
	"harboured", "harbored",
	"harmonise", "harmonize",
	"honouring", "honoring",
	"humanised", "humanized",
	"humanises", "humanizes",
	"humouring", "humoring",
	"hybridise", "hybridize",
	"hypnotise", "hypnotize",
	"idealised", "idealized",
	"idealises", "idealizes",
	"idolising", "idolizing",
	"immunised", "immunized",
	"immunises", "immunizes",
	"inflexion", "inflection",
	"italicise", "italicize",
	"itemising", "itemizing",
	"jewellers", "jewelers",
	"jewellery", "jewelry",
	"kilometre", "kilometer",
	"labelling", "labeling",
	"labourers", "laborers",
	"labouring", "laboring",
	"legalised", "legalized",
	"legalises", "legalizes",
	"leukaemia", "leukemia",
	"levellers", "levelers",
	"levelling", "leveling",
	"libelling", "libeling",
	"libellous", "libelous",
	"licencing", "licensing",
	"lionising", "lionizing",
	"liquidise", "liquidize",
	"localised", "localized",
	"localises", "localizes",
	"magnetise", "magnetize",
	"manoeuvre", "maneuver",
	"marvelled", "marveled",
	"maximised", "maximized",
	"maximises", "maximizes",
	"mechanise", "mechanize",
	"mediaeval", "medieval",
	"memorised", "memorized",
	"memorises", "memorizes",
	"mesmerise", "mesmerize",
	"minimised", "minimized",
	"minimises", "minimizes",
	"mobilised", "mobilized",
	"mobilises", "mobilizes",
	"modellers", "modelers",
	"modelling", "modeling",
	"modernise", "modernize",
	"moralised", "moralized",
	"moralises", "moralizes",
	"motorised", "motorized",
	"mouldered", "moldered",
	"mouldiest", "moldiest",
	"mouldings", "moldings",
	"moustache", "mustache",
	"neighbour", "neighbor",
	"normalise", "normalize",
	"odourless", "odorless",
	"oestrogen", "estrogen",
	"optimised", "optimized",
	"optimises", "optimizes",
	"organised", "organized",
	"organiser", "organizer",
	"organises", "organizes",
	"ostracise", "ostracize",
	"oxidising", "oxidizing",
	"paederast", "pederast",
	"panelling", "paneling",
	"panellist", "panelist",
	"paralysed", "paralyzed",
	"paralyses", "paralyzes",
	"parcelled", "parceled",
	"passivise", "passivize",
	"patronise", "patronize",
	"pedalling", "pedaling",
	"penalised", "penalized",
	"penalises", "penalizes",
	"pencilled", "penciled",
	"ploughing", "plowing",
	"ploughman", "plowman",
	"ploughmen", "plowmen",
	"polarised", "polarized",
	"polarises", "polarizes",
	"practised", "practiced",
	"practises", "practices",
	"pretences", "pretenses",
	"primaeval", "primeval",
	"privatise", "privatize",
	"programme", "program",
	"publicise", "publicize",
	"pulverise", "pulverize",
	"pummelled", "pummel",
	"randomise", "randomize",
	"ravelling", "raveling",
	"realising", "realizing",
	"recognise", "recognize",
	"refuelled", "refueled",
	"remoulded", "remolded",
	"revellers", "revelers",
	"revelling", "reveling",
	"rivalling", "rivaling",
	"saltpetre", "saltpeter",
	"sanitised", "sanitized",
	"sanitises", "sanitizes",
	"satirised", "satirized",
	"satirises", "satirizes",
	"savouries", "savories",
	"savouring", "savoring",
	"sceptical", "skeptical",
	"sensitise", "sensitize",
	"sepulchre", "sepulcher",
	"serialise", "serialize",
	"sermonise", "sermonize",
	"shovelled", "shoveled",
	"signalise", "signalize",
	"signalled", "signaled",
	"snivelled", "sniveled",
	"socialise", "socialize",
	"sodomised", "sodomized",
	"sodomises", "sodomizes",
	"solemnise", "solemnize",
	"spiralled", "spiraled",
	"splendour", "splendor",
	"stabilise", "stabilize",
	"sterilise", "sterilize",
	"subsidise", "subsidize",
	"succoured", "succored",
	"sulphates", "sulfates",
	"sulphides", "sulfides",
	"summarise", "summarize",
	"swivelled", "swiveled",
	"symbolise", "symbolize",
	"syphoning", "siphoning",
	"tantalise", "tantalize",
	"tasselled", "tasseled",
	"temporise", "temporize",
	"tenderise", "tenderize",
	"terrorise", "terrorize",
	"theorised", "theorized",
	"theorises", "theorizes",
	"towelling", "toweling",
	"travelled", "traveled",
	"traveller", "traveler",
	"trialling", "trialing",
	"tricolour", "tricolor",
	"tunnelled", "tunneled",
	"tyrannise", "tyrannize",
	"unionised", "unionized",
	"unionises", "unionizes",
	"unsavoury", "unsavory",
	"urbanised", "urbanized",
	"urbanises", "urbanizes",
	"utilising", "utilizing",
	"vandalise", "vandalize",
	"vaporised", "vaporized",
	"vaporises", "vaporizes",
	"verbalise", "verbalize",
	"victimise", "victimize",
	"visualise", "visualize",
	"vocalised", "vocalized",
	"vocalises", "vocalizes",
	"vulgarise", "vulgarize",
	"weaselled", "weaseled",
	"womanised", "womanized",
	"womaniser", "womanizer",
	"womanises", "womanizes",
	"yodelling", "yodeling",
	"yoghourts", "yogurts",
	"agonised", "agonized",
	"agonises", "agonizes",
	"almanack", "almanac",
	"amortise", "amortize",
	"analogue", "analog",
	"analysed", "analyzed",
	"armoured", "armored",
	"armourer", "armorer",
	"artefact", "artifact",
	"baptised", "baptized",
	"baptises", "baptizes",
	"baulking", "balking",
	"belabour", "belabor",
	"bevelled", "beveled",
	"calibres", "calibers",
	"calliper", "caliper",
	"canalise", "canalize",
	"canonise", "canonize",
	"carolled", "caroled",
	"catalyse", "catalyze",
	"cavilled", "caviled",
	"civilise", "civilize",
	"clamours", "clamors",
	"clangour", "clangor",
	"colonise", "colonize",
	"coloured", "colored",
	"cosiness", "coziness",
	"crueller", "crueler",
	"defences", "defenses",
	"demonise", "demonize",
	"deputise", "deputize",
	"dialling", "dialing",
	"dialogue", "dialog",
	"digitise", "digitize",
	"draughty", "drafty",
	"duelling", "dueling",
	"energise", "energize",
	"enthrals", "enthralls",
	"equalise", "equalize",
	"eulogise", "eulogize",
	"favoured", "favored",
	"feminise", "feminize",
	"finalise", "finalize",
	"flautist", "flutist",
	"flavours", "flavors",
	"foetuses", "fetuses",
	"fuelling", "fueling",
	"gaolbird", "jailbird",
	"gryphons", "griffins",
	"harbours", "harbors",
	"honoured", "honored",
	"humanise", "humanize",
	"humoured", "humored",
	"idealise", "idealize",
	"idolised", "idolized",
	"idolises", "idolizes",
	"immunise", "immunize",
	"ionisers", "ionizers",
	"ionising", "ionizing",
	"itemised", "itemized",
	"itemises", "itemizes",
	"jewelled", "jeweled",
	"jeweller", "jeweler",
	"labelled", "labeled",
	"laboured", "labored",
	"labourer", "laborer",
	"legalise", "legalize",
	"levelled", "leveled",
	"leveller", "leveler",
	"libelled", "libeled",
	"licenced", "licensed",
	"licences", "licenses",
	"lionised", "lionized",
	"lionises", "lionizes",
	"localise", "localize",
	"maximise", "maximize",
	"memorise", "memorize",
	"minimise", "minimize",
	"misspelt", "misspelled",
	"mobilise", "mobilize",
	"modelled", "modeled",
	"modeller", "modeler",
	"moralise", "moralize",
	"moulders", "molders",
	"mouldier", "moldier",
	"moulding", "molding",
	"moulting", "molting",
	"offences", "offenses",
	"optimise", "optimize",
	"organise", "organize",
	"oxidised", "oxidized",
	"oxidises", "oxidizes",
	"panelled", "paneled",
	"paralyse", "paralyze",
	"parlours", "parlors",
	"pedalled", "pedaled",
	"penalise", "penalize",
	"philtres", "filters",
	"ploughed", "plowed",
	"polarise", "polarize",
	"practise", "practice",
	"pretence", "pretense",
	"ravelled", "raveled",
	"realised", "realized",
	"realises", "realizes",
	"remoulds", "remolds",
	"revelled", "reveled",
	"reveller", "reveler",
	"rivalled", "rivaled",
	"rumoured", "rumored",
	"sanitise", "sanitize",
	"satirise", "satirize",
	"saviours", "saviors",
	"savoured", "savored",
	"sceptics", "skeptics",
	"sceptres", "scepters",
	"sodomise", "sodomize",
	"spectres", "specters",
	"succours", "succors",
	"sulphate", "sulfate",
	"sulphide", "sulfide",
	"syphoned", "siphoned",
	"theatres", "theaters",
	"theorise", "theorize",
	"towelled", "toweled",
	"toxaemia", "toxemia",
	"trialled", "trialed",
	"unionise", "unionize",
	"urbanise", "urbanize",
	"utilised", "utilized",
	"utilises", "utilizes",
	"vaporise", "vaporize",
	"vocalise", "vocalize",
	"womanise", "womanize",
	"yodelled", "yodeled",
	"yoghourt", "yogurt",
	"yoghurts", "yogurts",
	"agonise", "agonize",
	"anaemia", "anemia",
	"anaemic", "anemic",
	"analyse", "analyze",
	"arbours", "arbors",
	"armoury", "armory",
	"baptise", "baptize",
	"baulked", "balked",
	"behoved", "behooved",
	"behoves", "behooves",
	"calibre", "caliber",
	"candour", "candor",
	"centred", "centered",
	"centres", "centers",
	"cheques", "checks",
	"clamour", "clamor",
	"colours", "colors",
	"cosiest", "coziest",
	"defence", "defense",
	"dialled", "dialed",
	"distils", "distills",
	"duelled", "dueled",
	"enthral", "enthrall",
	"favours", "favors",
	"fervour", "fervor",
	"flavour", "flavor",
	"fuelled", "fueled",
	"fulfils", "fulfills",
	"gaolers", "jailers",
	"gaoling", "jailing",
	"gipsies", "gypsies",
	"glueing", "gluing",
	"goitres", "goiters",
	"grammes", "grams",
	"groynes", "groins",
	"gryphon", "griffin",
	"harbour", "harbor",
	"honours", "honors",
	"humours", "humors",
	"idolise", "idolize",
	"instals", "installs",
	"instils", "instills",
	"ionised", "ionized",
	"ioniser", "ionizer",
	"ionises", "ionizes",
	"itemise", "itemize",
	"labours", "labors",
	"licence", "license",
	"lionise", "lionize",
	"louvred", "louvered",
	"louvres", "louvers",
	"moulded", "molded",
	"moulder", "molder",
	"moulted", "molted",
	"offence", "offense",
	"oxidise", "oxidize",
	"parlour", "parlor",
	"philtre", "filter",
	"ploughs", "plows",
	"pyjamas", "pajamas",
	"rancour", "rancor",
	"realise", "realize",
	"remould", "remold",
	"rigours", "rigors",
	"rumours", "rumors",
	"saviour", "savior",
	"savours", "savors",
	"savoury", "savory",
	"sceptic", "skeptic",
	"sceptre", "scepter",
	"spectre", "specter",
	"storeys", "stories",
	"succour", "succor",
	"sulphur", "sulfur",
	"syphons", "siphons",
	"theatre", "theater",
	"tumours", "tumors",
	"utilise", "utilize",
	"vapours", "vapors",
	"waggons", "wagons",
	"yoghurt", "yogurt",
	"ageing", "aging",
	"appals", "appalls",
	"arbour", "arbor",
	"ardour", "ardor",
	"baulks", "balks",
	"behove", "behoove",
	"centre", "center",
	"cheque", "check",
	"chilli", "chili",
	"colour", "color",
	"cosier", "cozier",
	"cosies", "cozies",
	"cosily", "cozily",
	"distil", "distill",
	"edoema", "edema",
	"enrols", "enrolls",
	"faecal", "fecal",
	"faeces", "feces",
	"favour", "favor",
	"fibres", "fibers",
	"foetal", "fetal",
	"foetid", "fetid",
	"foetus", "fetus",
	"fulfil", "fulfill",
	"gaoled", "jailed",
	"gaoler", "jailer",
	"goitre", "goiter",
	"gramme", "gram",
	"groyne", "groin",
	"honour", "honor",
	"humour", "humor",
	"instal", "install",
	"instil", "instill",
	"ionise", "ionize",
	"labour", "labor",
	"litres", "liters",
	"lustre", "luster",
	"meagre", "meager",
	"metres", "meters",
	"mitres", "miters",
	"moulds", "molds",
	"mouldy", "moldy",
	"moults", "molts",
	"odours", "odors",
	"plough", "plow",
	"pyjama", "pajama",
	"rigour", "rigor",
	"rumour", "rumor",
	"savour", "savor",
	"storey", "story",
	"syphon", "siphon",
	"tumour", "tumor",
	"valour", "valor",
	"vapour", "vapor",
	"vigour", "vigor",
	"waggon", "wagon",
	"appal", "appall",
	"baulk", "balk",
	"enrol", "enroll",
	"fibre", "fiber",
	"gaols", "jails",
	"litre", "liter",
	"metre", "meter",
	"mitre", "miter",
	"mould", "mold",
	"moult", "molt",
	"odour", "odor",
	"tyres", "tires",
	"cosy", "cozy",
	"gaol", "jail",
	"tyre", "tire",
}
