// Code generated by initialismer; DO NOT EDIT.
package canonicalheader

// initialism mapping of not canonical headers from
// https://en.wikipedia.org/wiki/List_of_HTTP_header_fields
// https://www.iana.org/assignments/http-fields/http-fields.xhtml.
func initialism() map[string]string {
	return map[string]string{
		"A-Im":                              "A-IM",
		"Accept-Ch":                         "Accept-CH",
		"Alpn":                              "ALPN",
		"Amp-Cache-Transform":               "AMP-Cache-Transform",
		"C-Pep":                             "C-PEP",
		"C-Pep-Info":                        "C-PEP-Info",
		"Cal-Managed-Id":                    "Cal-Managed-ID",
		"Caldav-Timezones":                  "CalDAV-Timezones",
		"Cdn-Cache-Control":                 "CDN-Cache-Control",
		"Cdn-Loop":                          "CDN-Loop",
		"Content-Id":                        "Content-ID",
		"Content-Md5":                       "Content-MD5",
		"Dasl":                              "DASL",
		"Dav":                               "DAV",
		"Differential-Id":                   "Differential-ID",
		"Dnt":                               "DNT",
		"Dpop":                              "DPoP",
		"Dpop-Nonce":                        "DPoP-Nonce",
		"Ediint-Features":                   "EDIINT-Features",
		"Etag":                              "ETag",
		"Expect-Ct":                         "Expect-CT",
		"Getprofile":                        "GetProfile",
		"Http2-Settings":                    "HTTP2-Settings",
		"Im":                                "IM",
		"Include-Referred-Token-Binding-Id": "Include-Referred-Token-Binding-ID",
		"Last-Event-Id":                     "Last-Event-ID",
		"Mime-Version":                      "MIME-Version",
		"Nel":                               "NEL",
		"Odata-Entityid":                    "OData-EntityId",
		"Odata-Isolation":                   "OData-Isolation",
		"Odata-Maxversion":                  "OData-MaxVersion",
		"Odata-Version":                     "OData-Version",
		"Optional-Www-Authenticate":         "Optional-WWW-Authenticate",
		"Oscore":                            "OSCORE",
		"Oslc-Core-Version":                 "OSLC-Core-Version",
		"P3p":                               "P3P",
		"Pep":                               "PEP",
		"Pep-Info":                          "PEP-Info",
		"Pics-Label":                        "PICS-Label",
		"Profileobject":                     "ProfileObject",
		"Repeatability-Client-Id":           "Repeatability-Client-ID",
		"Repeatability-Request-Id":          "Repeatability-Request-ID",
		"Sec-Gpc":                           "Sec-GPC",
		"Sec-Websocket-Accept":              "Sec-WebSocket-Accept",
		"Sec-Websocket-Extensions":          "Sec-WebSocket-Extensions",
		"Sec-Websocket-Key":                 "Sec-WebSocket-Key",
		"Sec-Websocket-Protocol":            "Sec-WebSocket-Protocol",
		"Sec-Websocket-Version":             "Sec-WebSocket-Version",
		"Setprofile":                        "SetProfile",
		"Slug":                              "SLUG",
		"Soapaction":                        "SoapAction",
		"Status-Uri":                        "Status-URI",
		"Tcn":                               "TCN",
		"Te":                                "TE",
		"Ttl":                               "TTL",
		"Uri":                               "URI",
		"Www-Authenticate":                  "WWW-Authenticate",
		"X-Correlation-Id":                  "X-Correlation-ID",
		"X-Dns-Prefetch-Control":            "X-DNS-Prefetch-Control",
		"X-Real-Ip":                         "X-Real-IP",
		"X-Request-Id":                      "X-Request-ID",
		"X-Ua-Compatible":                   "X-UA-Compatible",
		"X-Webkit-Csp":                      "X-WebKit-CSP",
		"X-Xss":                             "X-XSS",
		"X-Xss-Protection":                  "X-XSS-Protection",
	}
}
