// Code generated by "script/gen.go"; DO NOT EDIT.

package stdinfo

var generatedPathByName = map[string]string{
	"fmt":       "fmt",                 // Freq=15795
	"testing":   "testing",             // Freq=12807
	"context":   "context",             // Freq=10797
	"time":      "time",                // Freq=8900
	"strings":   "strings",             // Freq=8852
	"os":        "os",                  // Freq=5712
	"bytes":     "bytes",               // Freq=4129
	"io":        "io",                  // Freq=3981
	"http":      "net/http",            // Freq=3691
	"sync":      "sync",                // Freq=3492
	"errors":    "errors",              // Freq=3107
	"strconv":   "strconv",             // Freq=3076
	"reflect":   "reflect",             // Freq=3025
	"filepath":  "path/filepath",       // Freq=2843
	"json":      "encoding/json",       // Freq=2537
	"sort":      "sort",                // Freq=2382
	"ioutil":    "io/ioutil",           // Freq=2164
	"net":       "net",                 // Freq=2025
	"math":      "math",                // Freq=1746
	"url":       "net/url",             // Freq=1411
	"regexp":    "regexp",              // Freq=1320
	"runtime":   "runtime",             // Freq=1318
	"log":       "log",                 // Freq=1149
	"flag":      "flag",                // Freq=1002
	"path":      "path",                // Freq=993
	"unsafe":    "unsafe",              // Freq=992
	"rand":      "math/rand",           // Freq=981
	"syscall":   "syscall",             // Freq=902
	"atomic":    "sync/atomic",         // Freq=804
	"bufio":     "bufio",               // Freq=695
	"httptest":  "net/http/httptest",   // Freq=676
	"exec":      "os/exec",             // Freq=676
	"binary":    "encoding/binary",     // Freq=476
	"tls":       "crypto/tls",          // Freq=475
	"token":     "go/token",            // Freq=471
	"utf8":      "unicode/utf8",        // Freq=404
	"base64":    "encoding/base64",     // Freq=383
	"ast":       "go/ast",              // Freq=373
	"x509":      "crypto/x509",         // Freq=357
	"hex":       "encoding/hex",        // Freq=340
	"unicode":   "unicode",             // Freq=309
	"types":     "go/types",            // Freq=309
	"big":       "math/big",            // Freq=230
	"sha256":    "crypto/sha256",       // Freq=227
	"template":  "text/template",       // Freq=211
	"fs":        "io/fs",               // Freq=162
	"parser":    "go/parser",           // Freq=160
	"sql":       "database/sql",        // Freq=157
	"gzip":      "compress/gzip",       // Freq=150
	"signal":    "os/signal",           // Freq=139
	"pem":       "encoding/pem",        // Freq=137
	"hash":      "hash",                // Freq=137
	"crypto":    "crypto",              // Freq=132
	"build":     "go/build",            // Freq=121
	"debug":     "runtime/debug",       // Freq=121
	"bits":      "math/bits",           // Freq=120
	"constant":  "go/constant",         // Freq=120
	"xml":       "encoding/xml",        // Freq=118
	"tabwriter": "text/tabwriter",      // Freq=116
	"md5":       "crypto/md5",          // Freq=110
	"rsa":       "crypto/rsa",          // Freq=103
	"format":    "go/format",           // Freq=88
	"sha1":      "crypto/sha1",         // Freq=85
	"driver":    "database/sql/driver", // Freq=81
	"pkix":      "crypto/x509/pkix",    // Freq=80
	"heap":      "container/heap",      // Freq=78
	"tar":       "archive/tar",         // Freq=77
	"ecdsa":     "crypto/ecdsa",        // Freq=75
	"cipher":    "crypto/cipher",       // Freq=74
	"crc32":     "hash/crc32",          // Freq=70
	"gob":       "encoding/gob",        // Freq=65
	"elliptic":  "crypto/elliptic",     // Freq=60
	"subtle":    "crypto/subtle",       // Freq=54
	"zip":       "archive/zip",         // Freq=54
	"aes":       "crypto/aes",          // Freq=53
	"mime":      "mime",                // Freq=51
	"pprof":     "runtime/pprof",       // Freq=47
	"textproto": "net/textproto",       // Freq=47
	"image":     "image",               // Freq=45
	"fnv":       "hash/fnv",            // Freq=45
	"hmac":      "crypto/hmac",         // Freq=45
	"httputil":  "net/http/httputil",   // Freq=44
	"elf":       "debug/elf",           // Freq=44
	"encoding":  "encoding",            // Freq=41
	"sha512":    "crypto/sha512",       // Freq=41
	"cmplx":     "math/cmplx",          // Freq=40
	"color":     "image/color",         // Freq=38
	"html":      "html",                // Freq=37
	"expvar":    "expvar",              // Freq=34
	"embed":     "embed",               // Freq=32
	"csv":       "encoding/csv",        // Freq=31
	"importer":  "go/importer",         // Freq=31
	"multipart": "mime/multipart",      // Freq=30
	"printer":   "go/printer",          // Freq=27
	"syslog":    "log/syslog",          // Freq=27
	"asn1":      "encoding/asn1",       // Freq=27
	"list":      "container/list",      // Freq=27
	"scanner":   "go/scanner",          // Freq=25
	"ed25519":   "crypto/ed25519",      // Freq=25
	"dwarf":     "debug/dwarf",         // Freq=23
	"flate":     "compress/flate",      // Freq=22
	"zlib":      "compress/zlib",       // Freq=21
	"png":       "image/png",           // Freq=20
	"trace":     "runtime/trace",       // Freq=20
	"httptrace": "net/http/httptrace",  // Freq=19
	"utf16":     "unicode/utf16",       // Freq=19
	"rpc":       "net/rpc",             // Freq=19
	"macho":     "debug/macho",         // Freq=16
	"iotest":    "testing/iotest",      // Freq=15
	"dsa":       "crypto/dsa",          // Freq=13
	"parse":     "text/template/parse", // Freq=13
	"cookiejar": "net/http/cookiejar",  // Freq=12
	"fstest":    "testing/fstest",      // Freq=11
	"jpeg":      "image/jpeg",          // Freq=11
}

var generatedPackagesList = []Package{
	{Name: "adler32", Path: "hash/adler32", Freq: 7},
	{Name: "aes", Path: "crypto/aes", Freq: 53},
	{Name: "ascii85", Path: "encoding/ascii85", Freq: -1},
	{Name: "asn1", Path: "encoding/asn1", Freq: 27},
	{Name: "ast", Path: "go/ast", Freq: 373},
	{Name: "atomic", Path: "sync/atomic", Freq: 804},
	{Name: "base32", Path: "encoding/base32", Freq: 5},
	{Name: "base64", Path: "encoding/base64", Freq: 383},
	{Name: "big", Path: "math/big", Freq: 230},
	{Name: "binary", Path: "encoding/binary", Freq: 476},
	{Name: "bits", Path: "math/bits", Freq: 120},
	{Name: "bufio", Path: "bufio", Freq: 695},
	{Name: "build", Path: "go/build", Freq: 121},
	{Name: "bytes", Path: "bytes", Freq: 4129},
	{Name: "bzip2", Path: "compress/bzip2", Freq: 7},
	{Name: "cgi", Path: "net/http/cgi", Freq: 1},
	{Name: "cgo", Path: "runtime/cgo", Freq: -1},
	{Name: "cipher", Path: "crypto/cipher", Freq: 74},
	{Name: "cmplx", Path: "math/cmplx", Freq: 40},
	{Name: "color", Path: "image/color", Freq: 38},
	{Name: "constant", Path: "go/constant", Freq: 120},
	{Name: "constraint", Path: "go/build/constraint", Freq: 5},
	{Name: "context", Path: "context", Freq: 10797},
	{Name: "cookiejar", Path: "net/http/cookiejar", Freq: 12},
	{Name: "crc32", Path: "hash/crc32", Freq: 70},
	{Name: "crc64", Path: "hash/crc64", Freq: 3},
	{Name: "crypto", Path: "crypto", Freq: 132},
	{Name: "csv", Path: "encoding/csv", Freq: 31},
	{Name: "debug", Path: "runtime/debug", Freq: 121},
	{Name: "des", Path: "crypto/des", Freq: 8},
	{Name: "doc", Path: "go/doc", Freq: 15},
	{Name: "draw", Path: "image/draw", Freq: 7},
	{Name: "driver", Path: "database/sql/driver", Freq: 81},
	{Name: "dsa", Path: "crypto/dsa", Freq: 13},
	{Name: "dwarf", Path: "debug/dwarf", Freq: 23},
	{Name: "ecdsa", Path: "crypto/ecdsa", Freq: 75},
	{Name: "ed25519", Path: "crypto/ed25519", Freq: 25},
	{Name: "elf", Path: "debug/elf", Freq: 44},
	{Name: "elliptic", Path: "crypto/elliptic", Freq: 60},
	{Name: "embed", Path: "embed", Freq: 32},
	{Name: "encoding", Path: "encoding", Freq: 41},
	{Name: "errors", Path: "errors", Freq: 3107},
	{Name: "exec", Path: "os/exec", Freq: 676},
	{Name: "expvar", Path: "expvar", Freq: 34},
	{Name: "fcgi", Path: "net/http/fcgi", Freq: 2},
	{Name: "filepath", Path: "path/filepath", Freq: 2843},
	{Name: "flag", Path: "flag", Freq: 1002},
	{Name: "flate", Path: "compress/flate", Freq: 22},
	{Name: "fmt", Path: "fmt", Freq: 15795},
	{Name: "fnv", Path: "hash/fnv", Freq: 45},
	{Name: "format", Path: "go/format", Freq: 88},
	{Name: "fs", Path: "io/fs", Freq: 162},
	{Name: "fstest", Path: "testing/fstest", Freq: 11},
	{Name: "gif", Path: "image/gif", Freq: 5},
	{Name: "gob", Path: "encoding/gob", Freq: 65},
	{Name: "gosym", Path: "debug/gosym", Freq: 3},
	{Name: "gzip", Path: "compress/gzip", Freq: 150},
	{Name: "hash", Path: "hash", Freq: 137},
	{Name: "heap", Path: "container/heap", Freq: 78},
	{Name: "hex", Path: "encoding/hex", Freq: 340},
	{Name: "hmac", Path: "crypto/hmac", Freq: 45},
	{Name: "html", Path: "html", Freq: 37},
	{Name: "http", Path: "net/http", Freq: 3691},
	{Name: "httptest", Path: "net/http/httptest", Freq: 676},
	{Name: "httptrace", Path: "net/http/httptrace", Freq: 19},
	{Name: "httputil", Path: "net/http/httputil", Freq: 44},
	{Name: "image", Path: "image", Freq: 45},
	{Name: "importer", Path: "go/importer", Freq: 31},
	{Name: "io", Path: "io", Freq: 3981},
	{Name: "iotest", Path: "testing/iotest", Freq: 15},
	{Name: "ioutil", Path: "io/ioutil", Freq: 2164},
	{Name: "jpeg", Path: "image/jpeg", Freq: 11},
	{Name: "json", Path: "encoding/json", Freq: 2537},
	{Name: "jsonrpc", Path: "net/rpc/jsonrpc", Freq: -1},
	{Name: "list", Path: "container/list", Freq: 27},
	{Name: "log", Path: "log", Freq: 1149},
	{Name: "lzw", Path: "compress/lzw", Freq: 3},
	{Name: "macho", Path: "debug/macho", Freq: 16},
	{Name: "mail", Path: "net/mail", Freq: 7},
	{Name: "maphash", Path: "hash/maphash", Freq: 1},
	{Name: "math", Path: "math", Freq: 1746},
	{Name: "md5", Path: "crypto/md5", Freq: 110},
	{Name: "metrics", Path: "runtime/metrics", Freq: 3},
	{Name: "mime", Path: "mime", Freq: 51},
	{Name: "multipart", Path: "mime/multipart", Freq: 30},
	{Name: "net", Path: "net", Freq: 2025},
	{Name: "os", Path: "os", Freq: 5712},
	{Name: "palette", Path: "image/color/palette", Freq: 4},
	{Name: "parse", Path: "text/template/parse", Freq: 13},
	{Name: "parser", Path: "go/parser", Freq: 160},
	{Name: "path", Path: "path", Freq: 993},
	{Name: "pe", Path: "debug/pe", Freq: 12},
	{Name: "pem", Path: "encoding/pem", Freq: 137},
	{Name: "pkix", Path: "crypto/x509/pkix", Freq: 80},
	{Name: "plan9obj", Path: "debug/plan9obj", Freq: 1},
	{Name: "plugin", Path: "plugin", Freq: 4},
	{Name: "png", Path: "image/png", Freq: 20},
	{Name: "pprof", Path: "runtime/pprof", Freq: 47},
	{Name: "pprof", Path: "net/http/pprof", Freq: 33},
	{Name: "printer", Path: "go/printer", Freq: 27},
	{Name: "quick", Path: "testing/quick", Freq: 51},
	{Name: "quotedprintable", Path: "mime/quotedprintable", Freq: 2},
	{Name: "race", Path: "runtime/race", Freq: -1},
	{Name: "rand", Path: "math/rand", Freq: 981},
	{Name: "rand", Path: "crypto/rand", Freq: 256},
	{Name: "rc4", Path: "crypto/rc4", Freq: 3},
	{Name: "reflect", Path: "reflect", Freq: 3025},
	{Name: "regexp", Path: "regexp", Freq: 1320},
	{Name: "ring", Path: "container/ring", Freq: 2},
	{Name: "rpc", Path: "net/rpc", Freq: 19},
	{Name: "rsa", Path: "crypto/rsa", Freq: 103},
	{Name: "runtime", Path: "runtime", Freq: 1318},
	{Name: "scanner", Path: "text/scanner", Freq: 23},
	{Name: "scanner", Path: "go/scanner", Freq: 25},
	{Name: "sha1", Path: "crypto/sha1", Freq: 85},
	{Name: "sha256", Path: "crypto/sha256", Freq: 227},
	{Name: "sha512", Path: "crypto/sha512", Freq: 41},
	{Name: "signal", Path: "os/signal", Freq: 139},
	{Name: "smtp", Path: "net/smtp", Freq: 6},
	{Name: "sort", Path: "sort", Freq: 2382},
	{Name: "sql", Path: "database/sql", Freq: 157},
	{Name: "strconv", Path: "strconv", Freq: 3076},
	{Name: "strings", Path: "strings", Freq: 8852},
	{Name: "subtle", Path: "crypto/subtle", Freq: 54},
	{Name: "suffixarray", Path: "index/suffixarray", Freq: 2},
	{Name: "sync", Path: "sync", Freq: 3492},
	{Name: "syntax", Path: "regexp/syntax", Freq: 11},
	{Name: "syscall", Path: "syscall", Freq: 902},
	{Name: "syslog", Path: "log/syslog", Freq: 27},
	{Name: "tabwriter", Path: "text/tabwriter", Freq: 116},
	{Name: "tar", Path: "archive/tar", Freq: 77},
	{Name: "template", Path: "html/template", Freq: 173},
	{Name: "template", Path: "text/template", Freq: 211},
	{Name: "testing", Path: "testing", Freq: 12807},
	{Name: "textproto", Path: "net/textproto", Freq: 47},
	{Name: "time", Path: "time", Freq: 8900},
	{Name: "tls", Path: "crypto/tls", Freq: 475},
	{Name: "token", Path: "go/token", Freq: 471},
	{Name: "trace", Path: "runtime/trace", Freq: 20},
	{Name: "types", Path: "go/types", Freq: 309},
	{Name: "tzdata", Path: "time/tzdata", Freq: 6},
	{Name: "unicode", Path: "unicode", Freq: 309},
	{Name: "unsafe", Path: "unsafe", Freq: 992},
	{Name: "url", Path: "net/url", Freq: 1411},
	{Name: "user", Path: "os/user", Freq: 51},
	{Name: "utf16", Path: "unicode/utf16", Freq: 19},
	{Name: "utf8", Path: "unicode/utf8", Freq: 404},
	{Name: "x509", Path: "crypto/x509", Freq: 357},
	{Name: "xml", Path: "encoding/xml", Freq: 118},
	{Name: "zip", Path: "archive/zip", Freq: 54},
	{Name: "zlib", Path: "compress/zlib", Freq: 21},
}
