package dst

// notest

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ArrayType) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *AssignStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *BadDecl) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *BadExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *BadStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *BasicLit) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *BinaryExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *BlockStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *BranchStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *CallExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *CaseClause) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ChanType) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *CommClause) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *CompositeLit) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *DeclStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *DeferStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *Ellipsis) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *EmptyStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ExprStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *Field) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *FieldList) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *File) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ForStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *FuncDecl) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *FuncLit) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *FuncType) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *GenDecl) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *GoStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *Ident) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *IfStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ImportSpec) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *IncDecStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *IndexExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *IndexListExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *InterfaceType) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *KeyValueExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *LabeledStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *MapType) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations is nil for Package nodes.
func (n *Package) Decorations() *NodeDecs {
	return nil
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ParenExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *RangeStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ReturnStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *SelectStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *SelectorExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *SendStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *SliceExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *StarExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *StructType) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *SwitchStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *TypeAssertExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *TypeSpec) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *TypeSwitchStmt) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *UnaryExpr) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}

// Decorations returns the decorations that are common to all nodes (Before, Start, End, After).
func (n *ValueSpec) Decorations() *NodeDecs {
	return &n.Decs.NodeDecs
}
