package types

type Config struct {
	RequireSingleType bool // Require the use of a single global 'type' declaration only
	RequireGrouping   bool // Require the use of grouped global 'type' declarations
}
