//go:generate go run ../generate.go

// Package staticcheck contains analyzes that find bugs and performance issues.
// Barring the rare false positive, any code flagged by these analyzes needs to be fixed.
package staticcheck
